# Combo helpers independent of GUI framework - these operate on
# SelectionCallbackProperty objects.

import weakref

from glue.core import BaseData, Subset
from glue.core.hub import HubListener
from glue.core.message import (DataReorderComponentMessage,
                               ComponentsChangedMessage,
                               DataCollectionAddMessage,
                               DataCollectionDeleteMessage,
                               DataUpdateMessage,
                               DataRenameComponentMessage)
from echo import delay_callback, ChoiceSeparator

__all__ = ['ComponentIDComboHelper', 'ManualDataComboHelper',
           'DataCollectionComboHelper', 'ComboHelper', 'BaseDataComboHelper']


def unique_data_iter(datasets):
    """
    Return a list with only Data objects, with duplicates removed, but
    preserving the original order.
    """
    datasets_new = []
    for dataset in datasets:
        if isinstance(dataset, BaseData):
            if dataset not in datasets_new:
                datasets_new.append(dataset)
        else:
            if dataset.data not in datasets_new:
                datasets_new.append(dataset.data)
    return datasets_new


class ComboHelper(HubListener):
    """
    Base class for any combo helper represented by a SelectionCallbackProperty.

    This stores the state and selection property and exposes the ``state``,
    ``selection`` and ``choices`` properties.

    Parameters
    ----------
    state : :class:`~glue.core.state_objects.State`
        The state to which the selection property belongs
    selection_property : :class:`~echo.SelectionCallbackProperty`
        The selection property representing the combo.
    """

    def __init__(self, state, selection_property):

        self._state = weakref.ref(state)
        self.selection_property = selection_property

    @property
    def state(self):
        """
        The state to which the selection property belongs.
        """
        return self._state()

    @property
    def selection(self):
        """
        The current selected value.
        """
        return getattr(self.state, self.selection_property)

    @selection.setter
    def selection(self, selection):
        return setattr(self.state, self.selection_property, selection)

    @property
    def choices(self):
        """
        The current valid choices for the combo.
        """
        prop = getattr(type(self.state), self.selection_property)
        return prop.get_choices(self.state)

    @choices.setter
    def choices(self, choices):
        with delay_callback(self.state, self.selection_property):
            prop = getattr(type(self.state), self.selection_property)
            prop.set_choices(self.state, choices)

    @property
    def display(self):
        """
        The current display function for the combo (the function that relates
        the Python objects to the display label)
        """
        prop = getattr(type(self.state), self.selection_property)
        return prop.get_display_func(self.state)

    @display.setter
    def display(self, display):
        prop = getattr(type(self.state), self.selection_property)
        return prop.set_display_func(self.state, display)

    def _on_rename(self, msg):
        # If a component ID is renamed, we don't need to refresh because the
        # list of actual component IDs is the same as before. However, we do
        # need to trigger a refresh of any GUI combos that use this, so we
        # make the property notify a change. However, if we are inside a
        # delay_callback block, the property will not be enabled, and notify()
        # won't have any effect, in which case we set the 'force_next_sync'
        # option which means that when exiting from the delay_callback block,
        # this property will show up as having changed
        prop = getattr(type(self.state), self.selection_property)
        if prop.enabled(self.state):
            prop.notify(self.state, self.selection, self.selection)
        else:
            prop.force_next_sync(self.state)


class ComponentIDComboHelper(ComboHelper):
    """
    The purpose of this class is to set up a combo (represented by a
    SelectionCallbackProperty) showing componentIDs for one or more datasets, and to
    update these componentIDs if needed, for example if new components are added
    to a dataset, or if componentIDs are renamed. This is a GUI
    framework-independent implementation.

    Parameters
    ----------
    state : :class:`~glue.core.state_objects.State`
        The state to which the selection property belongs
    selection_property : :class:`~echo.SelectionCallbackProperty`
        The selection property representing the combo.
    data_collection : :class:`~glue.core.data_collection.DataCollection`, optional
        The data collection to which the datasets belong - if specified,
        this is used to remove datasets from the combo when they are removed
        from the data collection.
    data : :class:`~glue.core.data.Data`, optional
        If specified, set up the combo for this dataset only and don't allow
        datasets to be added/removed
    numeric : `bool`, optional
        Show numeric components
    datetime : `bool`, optional
        Show datetime components
    categorical : `bool`, optional
        Show categorical components
    pixel_coord : `bool`, optional
        Show pixel coordinate components
    world_coord : `bool`, optional
        Show world coordinate components
    derived : `bool`, optional
        Show derived components
    none : `bool` or `str`, optional
        Add an entry that means `None`. If a string, this is the display string
        that will be shown for the `None` entry, otherwise an empty string is
        shown.
    """

    def __init__(self, state, selection_property,
                 data_collection=None, data=None,
                 numeric=True, datetime=True, categorical=True,
                 pixel_coord=False, world_coord=False, derived=True, none=False):

        super(ComponentIDComboHelper, self).__init__(state, selection_property)

        if isinstance(none, str):
            self._none = True
            self._none_label = none
        else:
            self._none = none
            self._none_label = ''

        def display_func_label(cid):
            if cid is None:
                return self._none_label
            else:
                return cid.label

        self.display = display_func_label

        self._numeric = numeric
        self._datetime = datetime
        self._categorical = categorical
        self._pixel_coord = pixel_coord
        self._world_coord = world_coord
        self._derived = derived

        if data is None:
            self._manual_data = False
            self._data = []
        else:
            self._manual_data = True
            self._data = [data]

        self._data_collection = data_collection
        if data_collection is None:
            self.hub = None
        else:
            if data_collection.hub is None:
                raise ValueError("Hub on data collection is not set")
            else:
                self.hub = data_collection.hub

        if data is not None:
            self.refresh()

    def clear(self):
        self._data.clear()
        self.refresh()

    @property
    def numeric(self):
        return self._numeric

    @numeric.setter
    def numeric(self, value):
        self._numeric = value
        self.refresh()

    @property
    def datetime(self):
        return self._datetime

    @datetime.setter
    def datetime(self, value):
        self._datetime = value
        self.refresh()

    @property
    def categorical(self):
        return self._categorical

    @categorical.setter
    def categorical(self, value):
        self._categorical = value
        self.refresh()

    @property
    def pixel_coord(self):
        return self._pixel_coord

    @pixel_coord.setter
    def pixel_coord(self, value):
        self._pixel_coord = value
        self.refresh()

    @property
    def world_coord(self):
        return self._world_coord

    @world_coord.setter
    def world_coord(self, value):
        self._world_coord = value
        self.refresh()

    @property
    def derived(self):
        return self._derived

    @derived.setter
    def derived(self, value):
        self._derived = value
        self.refresh()

    @property
    def none(self):
        return self._none

    @none.setter
    def none(self, value):
        if isinstance(value, str):
            self._none = True
            self._none_label = value
        else:
            self._none = value
            self._none_label = ''
        self.refresh()

    def append_data(self, data, refresh=True):

        if self._manual_data:
            raise Exception("Cannot change data in ComponentIDComboHelper "
                            "initialized from a single dataset")

        if isinstance(data, Subset):
            data = data.data

        if self.hub is None:
            if data.hub is not None:
                self.hub = data.hub
        elif data.hub is not self.hub:
            raise ValueError("Data Hub is different from current hub")

        if data not in self._data:
            self._data.append(data)
            if refresh:
                self.refresh()

    def remove_data(self, data):

        if self._manual_data:
            raise Exception("Cannot change data in ComponentIDComboHelper "
                            "initialized from a single dataset")

        if data in self._data:
            self._data.remove(data)
            self.refresh()

    def set_multiple_data(self, datasets):
        """
        Add multiple datasets to the combo in one go (and clear any previous datasets).

        Parameters
        ----------
        datasets : `list`
            The list of :class:`~glue.core.data.Data` objects to add
        """
        if self._manual_data:
            raise Exception("Cannot change data in ComponentIDComboHelper "
                            "initialized from a single dataset")
        self._data.clear()
        for data in unique_data_iter(datasets):
            self.append_data(data, refresh=False)
        self.refresh()

    @property
    def hub(self):
        return self._hub

    @hub.setter
    def hub(self, value):
        self._hub = value
        if value is not None:
            self.register_to_hub(value)

    def refresh(self, *args):

        choices = []

        if self._none:
            choices.append(None)

        for data in self._data:

            derived_components = [cid for cid in data.derived_components if cid.parent is data]

            if len(self._data) > 1:
                if data.label is None or data.label == '':
                    choices.append(ChoiceSeparator('Untitled Data'))
                else:
                    choices.append(ChoiceSeparator(data.label))

            cids = [ChoiceSeparator('Main components')]
            for cid in data.main_components:
                if ((data.get_kind(cid) == 'numerical' and self.numeric) or
                        (data.get_kind(cid) == 'datetime' and self.datetime) or
                        (data.get_kind(cid) == 'categorical' and self.categorical)):
                    cids.append(cid)
            if len(cids) > 1:
                if self.pixel_coord or self.world_coord or (self.derived and len(derived_components) > 0):
                    choices += cids
                else:
                    choices += cids[1:]

            if self.numeric and self.derived:
                cids = [ChoiceSeparator('Derived components')]
                for cid in derived_components:
                    cids.append(cid)
                if len(cids) > 1:
                    choices += cids

            if self.pixel_coord or self.world_coord:
                cids = [ChoiceSeparator('Coordinate components')]
                if self.pixel_coord:
                    cids += data.pixel_component_ids
                if self.world_coord:
                    cids += data.world_component_ids
                if len(cids) > 1:
                    choices += cids

        self.choices = choices

    def _filter_msg(self, msg):
        return msg.sender in self._data

    def register_to_hub(self, hub):
        hub.subscribe(self, DataRenameComponentMessage,
                      handler=self._on_rename,
                      filter=self._filter_msg)
        hub.subscribe(self, DataReorderComponentMessage,
                      handler=self.refresh,
                      filter=self._filter_msg)
        hub.subscribe(self, ComponentsChangedMessage,
                      handler=self.refresh,
                      filter=self._filter_msg)
        if self._data_collection is not None:
            hub.subscribe(self, DataCollectionDeleteMessage,
                          handler=self._remove_data)

    def _remove_data(self, msg):
        self.remove_data(msg.data)

    def unregister(self, hub):
        hub.unsubscribe_all(self)


class BaseDataComboHelper(ComboHelper):
    """
    This is a base class for helpers for combo boxes that need to show a list
    of data objects.

    Parameters
    ----------
    state : :class:`~glue.core.state_objects.State`
        The state to which the selection property belongs
    selection_property : :class:`~echo.SelectionCallbackProperty`
        The selection property representing the combo.
    data_collection : :class:`~glue.core.data_collection.DataCollection`
        The data collection to which the datasets belong - this is needed
        because if a dataset is removed from the data collection, we want to
        remove it here.
    """

    def __init__(self, state, selection_property, data_collection=None):

        super(BaseDataComboHelper, self).__init__(state, selection_property)

        def display_func_label(cid):
            return cid.label

        self.display = display_func_label

        self._component_id_helpers = []
        self.state.add_callback(self.selection_property, self.refresh_component_ids)

        self._data_collection = data_collection

        if data_collection is not None:
            if data_collection.hub is None:
                raise ValueError("Hub on data collection is not set")
            else:
                self.hub = data_collection.hub
        else:
            self.hub = None

    def refresh(self, *args):
        self.choices = [data for data in self._datasets]
        self.refresh_component_ids()

    def refresh_component_ids(self, *args):
        data = getattr(self.state, self.selection_property)
        for helper in self._component_id_helpers:
            helper.clear()
            if data is not None:
                helper.append_data(data)
            helper.refresh()

    def add_component_id_combo(self, combo):
        helper = ComponentIDComboHelper(combo)
        self._component_id_helpers.append_data(helper)
        if self._data is not None:
            helper.append_data(self._data)

    @property
    def hub(self):
        return self._hub

    @hub.setter
    def hub(self, value):
        self._hub = value
        if value is not None:
            self.register_to_hub(value)

    def register_to_hub(self, hub):
        pass

    def _on_data_update(self, msg):
        if msg.attribute == 'label':
            self._on_rename(msg)
        else:
            self.refresh()


class ManualDataComboHelper(BaseDataComboHelper):
    """
    This is a helper for combo boxes that need to show a list of data objects
    that is manually curated.

    Datasets are added and removed using the
    :meth:`~ManualDataComboHelper.append_data` and
    :meth:`~ManualDataComboHelper.remove_data` methods.

    Parameters
    ----------
    state : :class:`~glue.core.state_objects.State`
        The state to which the selection property belongs
    selection_property : :class:`~echo.SelectionCallbackProperty`
        The selection property representing the combo.
    data_collection : :class:`~glue.core.data_collection.DataCollection`
        The data collection to which the datasets belong - this is needed
        because if a dataset is removed from the data collection, we want to
        remove it here.
    """

    def __init__(self, state, selection_property, data_collection=None):

        super(ManualDataComboHelper, self).__init__(state, selection_property,
                                                    data_collection=data_collection)

        self._datasets = []

    def set_multiple_data(self, datasets):
        """
        Add multiple datasets to the combo in one go (and clear any previous datasets).

        Parameters
        ----------
        datasets : `list`
            The list of :class:`~glue.core.data.Data` objects to add
        """
        self._datasets.clear()
        for data in unique_data_iter(datasets):
            self.append_data(data, refresh=False)
        self.refresh()

    def append_data(self, data, refresh=True):
        if data in self._datasets:
            return
        if self.hub is None and data.hub is not None:
            self.hub = data.hub
        self._datasets.append(data)
        if refresh:
            self.refresh()

    def remove_data(self, data):
        if data not in self._datasets:
            return
        self._datasets.remove(data)
        self.refresh()

    def _remove_data_msg(self, msg):
        self.remove_data(msg.data)

    def _filter_msg(self, msg):
        return msg.sender in self._datasets

    def _filter_msg_dc(self, msg):
        return msg.sender is self._data_collection

    def register_to_hub(self, hub):

        super(ManualDataComboHelper, self).register_to_hub(hub)

        hub.subscribe(self, DataUpdateMessage,
                      handler=self._on_data_update,
                      filter=self._filter_msg)
        hub.subscribe(self, DataCollectionDeleteMessage,
                      handler=self._remove_data_msg,
                      filter=self._filter_msg_dc)


class DataCollectionComboHelper(BaseDataComboHelper):
    """
    This is a helper for combo boxes that need to show a list of data objects
    that is always in sync with a :class:`~glue.core.data_collection.DataCollection`.

    Parameters
    ----------
    state : :class:`~glue.core.state_objects.State`
        The state to which the selection property belongs
    selection_property : :class:`~echo.SelectionCallbackProperty`
        The selection property representing the combo.
    data_collection : :class:`~glue.core.data_collection.DataCollection`
        The data collection with which to stay in sync
    """

    def __init__(self, state, selection_property, data_collection):

        super(DataCollectionComboHelper, self).__init__(state, selection_property,
                                                        data_collection=data_collection)

        self._datasets = data_collection

        self.refresh()

    def _filter_msg_in(self, msg):
        return msg.sender in self._datasets

    def _filter_msg_is(self, msg):
        return msg.sender is self._datasets

    def register_to_hub(self, hub):
        super(DataCollectionComboHelper, self).register_to_hub(hub)
        hub.subscribe(self, DataUpdateMessage,
                      handler=self._on_data_update,
                      filter=self._filter_msg_in)
        hub.subscribe(self, DataCollectionAddMessage,
                      handler=self.refresh,
                      filter=self._filter_msg_is)
        hub.subscribe(self, DataCollectionDeleteMessage,
                      handler=self.refresh,
                      filter=self._filter_msg_is)



"""
Glue's fitting classes are designed to be easily subclassed for performing
custom model fitting in Glue.
"""

import numpy as np

from glue.core.simpleforms import IntOption, Option

__all__ = ['BaseFitter1D',
           'PolynomialFitter',
           'AstropyFitter1D',
           'SimpleAstropyGaussianFitter',
           'BasicGaussianFitter']


class BaseFitter1D(object):

    """
    Base class for 1D fitters.

    This abstract class must be overwritten.
    """

    label = "Fitter"
    """A short label for the fit, used by the GUI"""

    param_names = []
    """list of parameter names that support restrictions"""

    def __init__(self, **params):
        self._constraints = {}

        for k, v in params.items():
            if k in self.param_names:
                self.set_constraint(k, value=v)
            else:
                setattr(self, k, v)

    def plot(self, fit_result, axes, x, linewidth=None, alpha=None, color=None, normalize=None):
        """
        Plot the result of a fit.

        :param fit_result: The output from fit
        :param axes: The Matplotlib axes to add the fit to
        :param x: The values of X at which to visualize the model

        :returns: A list of matplotlib artists. **This is important:**
                  plots will not be properly cleared if this isn't provided
        """
        y = self.predict(fit_result, x)
        if normalize is not None:
            y = normalize(y)
        result = axes.plot(x, y, color,
                           lw=linewidth, alpha=alpha,
                           scalex=False, scaley=False)
        return result

    def _sigma_to_weights(self, dy):
        if dy is not None:
            return 1. / np.asarray(dy) ** 2

    @property
    def options(self):
        """
        A dictionary of the current setting of each model hyperparameter.

        Hyperparameters are defined in subclasses by creating class-level
        :mod:`Option <glue.core.simpleforms>` attributes. This attribute
        dict maps ``{hyperparameter_name: current_value}``
        """
        result = []
        for typ in type(self).mro():
            result.extend(k for k, v in typ.__dict__.items()
                          if isinstance(v, Option))
        return dict((o, getattr(self, o)) for o in result)

    def summarize(self, fit_result, x, y, dy=None):
        """
        Return a textual summary of the fit.

        :param fit_result: The return value from :meth:`fit`
        :param x: The x values passed to :meth:`fit`
        :returns: A description of the fit result
        :rtype: str
        """
        return str(fit_result)

    @property
    def constraints(self):
        """
        A dict of the constraints on each parameter in :attr:`param_names`.
        Each value is itself a dict with 3 items:

        :key value: The default value
        :key fixed: True / False, indicating whether the parameter is fixed
        :key bounds: [min, max] or None, indicating lower/upper limits
        """
        result = {}
        for p in self.param_names:
            result[p] = dict(value=None, fixed=False, limits=None)
            result[p].update(self._constraints.get(p, {}))
        return result

    def set_constraint(self, parameter_name, value=None,
                       fixed=None, limits=None):
        """
        Update a constraint.

        :param parameter_name: name of the parameter to update
        :type parameter_name: str
        :param value: Set the default value (optional)
        :param limits: Set the limits to[min, max] (optional)
        :param fixed: Set whether the parameter is fixed (optional)
        """
        c = self._constraints.setdefault(parameter_name, {})
        if value is not None:
            c['value'] = value
        if fixed is not None:
            c['fixed'] = fixed
        if limits is not None:
            c['limits'] = limits

    def build_and_fit(self, x, y, dy=None):
        """
        Method which builds the arguments to fit, and calls that method
        """
        x = np.asarray(x).ravel()
        y = np.asarray(y).ravel()
        if dy is not None:
            dy = np.asarray(dy).ravel()

        return self.fit(x, y, dy=dy,
                        constraints=self.constraints,
                        **self.options)

    def fit(self, x, y, dy, constraints, **options):
        """
        Fit the model to data.

        *This must be overriden by a subclass.*

        :param x: The x values of the data
        :type x:  :class:`numpy.ndarray`
        :param y: The y values of the data
        :type y:  :class:`numpy.ndarray`
        :param dy: 1 sigma uncertainties on each datum (optional)
        :type dy: :class:`numpy.ndarray`
        :param constraints: The current value of the ``constraints`` property
        :param options: kwargs for model hyperparameters.

        :returns: An object representing the fit result.
        """

        raise NotImplementedError()

    def predict(self, fit_result, x):
        """
        Evaluate the model at a set of locations.

        **This must be overridden in a subclass.**

        :param fit_result: The result from the fit method
        :param x: Locations to evaluate model at
        :type x: :class:`numpy.ndarray`

        :returns: model(x)
        :rtype: :class:`numpy.ndarray`
        """
        raise NotImplementedError()


class AstropyFitter1D(BaseFitter1D):

    """
    A base class for wrapping :mod:`astropy.modeling`.

    Subclasses must override :attr:`model_cls` :attr:`fitting_cls`
    to point to the desired Astropy :mod:`model <astropy.modeling>`
    and :mod:`fitter <astropy.modeling.fitting>` classes.

    In addition, they should override :attr:`label` with a better label,
    and :meth:`parameter_guesses` to generate initial guesses
    """

    model_cls = None
    """class describing the model"""

    fitting_cls = None
    """class to fit the model"""

    label = "Base Astropy Fitter"
    """UI Label"""

    @property
    def param_names(self):
        return self.model_cls.param_names

    def predict(self, fit_result, x):
        model, _ = fit_result
        return model(x)

    def summarize(self, fit_result, x, y, dy=None):
        model, fitter = fit_result
        result = [_report_fitter(fitter), ""]
        pnames = list(sorted(model.param_names))
        maxlen = max(map(len, pnames))
        result.extend("%s = %e" % (p.ljust(maxlen), getattr(model, p).value)
                      for p in pnames)
        return "\n".join(result)

    def fit(self, x, y, dy, constraints):
        m, f = self._get_model_fitter(x, y, dy, constraints)

        dy = self._sigma_to_weights(dy)
        return f(m, x, y, weights=dy), f

    def _get_model_fitter(self, x, y, dy, constraints):
        if self.model_cls is None or self.fitting_cls is None:
            raise NotImplementedError("Model or fitting class is unspecified.")

        params = dict((k, v['value']) for k, v in constraints.items())

        # update unset parameters with guesses from data
        for k, v in self.parameter_guesses(x, y, dy).items():
            if params[k] is not None or constraints[k]['fixed']:
                continue
            params[k] = v

        m = self.model_cls(**params)
        f = self.fitting_cls()

        for param_name, constraint in constraints.items():
            param = getattr(m, param_name)
            if constraint['fixed']:
                param.fixed = True
            if constraint['limits']:
                param.min, param.max = constraint['limits']
        return m, f

    def parameter_guesses(self, x, y, dy):
        """
        Provide initial guesses for each model parameter.

        **The base implementation does nothing, and should be overridden**

        :param x: X - values of the data
        :type x: :class:`numpy.ndarray`
        :param y: Y - values of the data
        :type y: :class:`numpy.ndarray`
        :param dy: uncertainties on Y(assumed to be 1 sigma)
        :type dy: :class:`numpy.ndarray`

        :returns: A dict mapping ``{parameter_name: value guess}`` for each
                  parameter
        """
        return {}


def _gaussian_parameter_estimates(x, y, dy):

    amplitude = np.percentile(y, 95)
    y = np.maximum(y / y.sum(), 0)
    mean = (x * y).sum()
    stddev = np.sqrt((y * (x - mean) ** 2).sum())
    return dict(mean=mean, stddev=stddev, amplitude=amplitude)


class BasicGaussianFitter(BaseFitter1D):

    """
    Fallback Gaussian fitter, for astropy < 0.3.

    If :mod:`astropy.modeling` is installed, this class is replaced by
    :class:`SimpleAstropyGaussianFitter`
    """
    label = "Gaussian"

    def _errorfunc(self, params, x, y, dy):
        yp = self.eval(x, *params)
        result = (yp - y)
        if dy is not None:
            result /= dy
        return result

    @staticmethod
    def eval(x, amplitude, mean, stddev):
        return np.exp(-(x - mean) ** 2 / (2 * stddev ** 2)) * amplitude

    @staticmethod
    def fit_deriv(x, amplitude, mean, stddev):
        """
        Gaussian1D model function derivatives.
        """

        d_amplitude = np.exp(-0.5 / stddev ** 2 * (x - mean) ** 2)
        d_mean = amplitude * d_amplitude * (x - mean) / stddev ** 2
        d_stddev = amplitude * d_amplitude * (x - mean) ** 2 / stddev ** 3
        return [d_amplitude, d_mean, d_stddev]

    def fit(self, x, y, dy, constraints):
        from scipy import optimize
        init_values = _gaussian_parameter_estimates(x, y, dy)
        init_values = [init_values[p] for p in ['amplitude', 'mean', 'stddev']]
        farg = (x, y, dy)
        dfunc = None
        fitparams, status, dinfo, mess, ierr = optimize.leastsq(
            self._errorfunc, init_values, args=farg, Dfun=dfunc,
            full_output=True)
        return fitparams

    def predict(self, fit_result, x):
        return self.eval(x, *fit_result)

    def summarize(self, fit_result, x, y, dy=None):
        return ("amplitude = %e\n"
                "mean      = %e\n"
                "stddev    = %e" % tuple(fit_result))


GaussianFitter = BasicGaussianFitter


try:
    from astropy.modeling import models, fitting

    class SimpleAstropyGaussianFitter(AstropyFitter1D):

        """
        Gaussian fitter using astropy.modeling.
        """
        model_cls = models.Gaussian1D
        try:
            fitting_cls = fitting.LevMarLSQFitter
        except AttributeError:  # astropy v0.3
            fitting_cls = fitting.NonLinearLSQFitter

        label = "Gaussian"

        parameter_guesses = staticmethod(_gaussian_parameter_estimates)

    GaussianFitter = SimpleAstropyGaussianFitter

except ImportError:
    pass


class PolynomialFitter(BaseFitter1D):

    """
    A polynomial model.

    The degree of the polynomial is specified by :attr:`degree`
    """
    label = "Polynomial"
    degree = IntOption(min=0, max=5, default=3, label="Polynomial Degree")

    def fit(self, x, y, dy, constraints, degree=2):
        """
        Fit a polynomial of order ``degree`` to the data.
        """
        w = self._sigma_to_weights(dy)

        return np.polyfit(x, y, degree, w=w)

    def predict(self, fit_result, x):
        return np.polyval(fit_result, x)

    def summarize(self, fit_result, x, y, dy=None):
        return "Coefficients:\n" + "\n".join("%e" % coeff
                                             for coeff in fit_result.tolist())


def _report_fitter(fitter):
    if "nfev" in fitter.fit_info:
        return "Converged in %i iterations" % fitter.fit_info['nfev']
    return 'Converged'


__FITTERS__ = [PolynomialFitter, GaussianFitter]



import numpy as np
import shapely

from glue.core.data import Data
from glue.core.contracts import contract

from glue.core.component import Component, ExtendedComponent


__all__ = ['RegionData']


class RegionData(Data):
    """
    A glue Data object for storing data that is associated with a region.

    This object can be used when a dataset describes 2D regions or 1D ranges. It
    contains exactly one :class:`~glue.core.component.ExtendedComponent` object
    which contains the boundaries of the regions, and must also contain
    one or two components that give the center of the regions in whatever data
    coordinates the regions are described in. Links in glue are not made
    directly on the :class:`~glue.core.component.ExtendedComponent`, but instead
    on the center components.

    Thus, a subset that includes the center of a region will include that region,
    but a subset that includes just a little part of the region will not include
    that region. These center components are not the same pixel components. For
    example, a dataset that is a table of 2D regions will have a single
    :class:`~glue.core.component.CoordinateComponent`, but must have two of these center
    components.

    A typical use case for this object is to store the properties of geographic
    regions, where the boundaries of the regions are stored in an
    :class:`~glue.core.component.ExtendedComponent` and the centers of the
    regions are stored in two components, one for the longitude and one for the
    latitude. Additional components may describe arbitrary properties of these
    geographic regions (e.g. population, area, etc).

    This class is mostly a convenience class. By using this class, Data
    Loaders can create RegionData directly from an iterable of geometries,
    since this class deals with creating representative points. Viewers
    can assume that when adding a RegionData object they are probably
    being asked to visualize the ExtendedComponent, and this class provides
    convenience methods for ascertaining whether the components currently
    visualized in a Viewer are the correct ones to enable display of the
    ExtendedComponent and to transform the regions through the glue
    linking architecture into the correct coordinates for display.

    Parameters
    ----------
    label : `str`, optional
        The label of the data.
    coords : :class:`~glue.core.coordinates.Coordinates`, optional
        The coordinates associated with the data.
    **kwargs
        All other keyword arguments are passed to the :class:`~glue.core.data.Data`
        constructor.

    Attributes
    ----------
    extended_component_id : :class:`~glue.core.component_id.ComponentID`
        The ID of the :class:`~glue.core.component.ExtendedComponent` object
        that contains the boundaries of the regions.
    center_x_id : :class:`~glue.core.component_id.ComponentID`
        The ID of the Component object that contains the x-coordinate of the
        center of the regions. This is actually stored in the component
        with the extended_component_id, but it is convenient to have it here.
    center_y_id : :class:`~glue.core.component_id.ComponentID`
        The ID of the Component object that contains the y-coordinate of the
        center of the regions. This is actually stored in the component
        with the extended_component_id, but it is convenient to have it here.

    Examples
    --------

    There are two main options for initializing a :class:`~glue.core.data_region.RegionData`
    object. The first is to simply pass in a list of ``Shapely.Geometry`` objects
    with dimesionality N, from which we will create N+1 components: one
    :class:`~glue.core.component.ExtendedComponent` with the boundaries, and N
    regular Component(s) with the center coordinates computed from the Shapley
    method :meth:`~shapely.GeometryCollection.representative_point`:

        >>> geometries = [shapely.geometry.Point(0, 0).buffer(1), shapely.geometry.Point(1, 1).buffer(1)]
        >>> my_region_data = RegionData(label='My Regions', boundary=geometries)

    This will create a :class:`~glue.core.data_region.RegionData` object with three
    components: one :class:`~glue.core.component.ExtendedComponent` with label
    "geo" and two regular Components with labels "Center [x] for boundary"
    and "Center [y] for boundary".

    The second is to explicitly create an :class:`~glue.core.component.ExtendedComponent`
    (which requires passing in the ComponentIDs for the center coordinates) and
    then use `add_component` to add this component to a :class:`~glue.core.data_region.RegionData`
    object. You might use this approach if your dataset already contains points that
    represent the centers of your regions and you want to avoid re-calculating them. For example:

        >>> center_x = [0, 1]
        >>> center_y = [0, 1]
        >>> geometries = [shapely.geometry.Point(0, 0).buffer(1), shapely.geometry.Point(1, 1).buffer(1)]

        >>> my_region_data = RegionData(label='My Regions')
        >>> # Region IDs are created and returned when we add a Component to a Data object
        >>> cen_x_id = my_region_data.add_component(center_x, label='Center [x]')
        >>> cen_y_id = my_region_data.add_component(center_y, label='Center [y]')
        >>> extended_comp = ExtendedComponent(geometries, center_comp_ids=[cen_x_id, cen_y_id])
        >>> my_region_data.add_component(extended_comp, label='boundaries')

    """

    def __init__(self, label="", coords=None, **kwargs):
        self._extended_component_id = None
        # __init__ calls add_component which deals with ExtendedComponent logic
        super().__init__(label=label, coords=coords, **kwargs)

    def __repr__(self):
        return f'RegionData (label: {self.label} | extended_component: {self.extended_component_id})'

    @property
    def center_x_id(self):
        return self.get_component(self.extended_component_id).x

    @property
    def center_y_id(self):
        return self.get_component(self.extended_component_id).y

    @property
    def extended_component_id(self):
        return self._extended_component_id

    @contract(component='component_like', label='cid_like')
    def add_component(self, component, label):
        """ Add a new component to this data set, allowing only one :class:`~glue.core.component.ExtendedComponent`

        If component is an array of Shapely objects then we use
        :meth:`~shapely.GeometryCollection.representative_point`: to
        create two new components for the center coordinates of the regions and
        add them to the :class:`~glue.core.data_region.RegionData` object as well.

        If component is an :class:`~glue.core.component.ExtendedComponent`,
        then we simply add it to the :class:`~glue.core.data_region.RegionData` object.

        We do this here instead of extending ``Component.autotyped`` because
        we only want to use :class:`~glue.core.component.ExtendedComponent` objects
        in the context of a :class:`~glue.core.data_region.RegionData` object.

        Parameters
        ----------
        component : :class:`~glue.core.component.Component` or array-like
            Object to add. If this is an array of Shapely objects, then we
            create two new components for the center coordinates of the regions
            as well.
        label : `str` or :class:`~glue.core.component_id.ComponentID`
              The label. If this is a string, a new
              :class:`glue.core.component_id.ComponentID`
              with this label will be created and associated with the Component.

        Raises
        ------
           `ValueError`, if the :class:`~glue.core.data_region.RegionData` already has an extended component
        """

        if not isinstance(component, Component):
            if all(isinstance(s, shapely.Geometry) for s in component):
                center_x = []
                center_y = []
                for s in component:
                    rep = s.representative_point()
                    center_x.append(rep.x)
                    center_y.append(rep.y)
                cen_x_id = super().add_component(np.asarray(center_x), f"Center [x] for {label}")
                cen_y_id = super().add_component(np.asarray(center_y), f"Center [y] for {label}")
                ext_component = ExtendedComponent(np.asarray(component), center_comp_ids=[cen_x_id, cen_y_id])
                self._extended_component_id = super().add_component(ext_component, label)
                return self._extended_component_id

        if isinstance(component, ExtendedComponent):
            if self.extended_component_id is not None:
                raise ValueError(f"Cannot add another ExtendedComponent; existing extended component: {self.extended_component_id}")
            else:
                self._extended_component_id = super().add_component(component, label)
                return self._extended_component_id
        else:
            return super().add_component(component, label)

    def _get_trans_to_cids(self, cen_cids, other_cids):
        """
        Use recursion to traverse links and build up a list of functions
        to convert cen_ids to other_cids.

        Parameters
        ----------
        cen_cids : list of :class:`~glue.core.component_id.ComponentID`
            The ComponentIDs that are the inputs to the transformation
            function.
        other_cids : list of :class:`~glue.core.component_id.ComponentID`
            The ComponentIDs that are the outputs of the transformation
            function.

        Raises
        ------
        ValueError
            If the links imply a transformation that cannot be done by Shapely.
        """
        if len(other_cids) != 2:
            raise ValueError("Can only deal with 2D transformations")
        linkx = self._get_external_link(other_cids[0])
        linky = self._get_external_link(other_cids[1])

        funcx = linkx.get_using()
        funcy = linky.get_using()

        if len(linkx.get_from_ids()) > 2 or len(linky.get_from_ids()) > 2:
            raise ValueError("Can only display regions if links depend on 2 or fewer other components.")

        def conv_function(x, y=None):
            if len(linkx.get_from_ids()) == 1 and len(linky.get_from_ids()) == 1:
                return [funcx(x), funcy(y)]
            else:
                return [funcx(x, y), funcy(x, y)]

        self.list_of_functions.append(conv_function)
        if len(linkx.get_from_ids()) == 2:
            other_cids = linkx.get_from_ids()
        else:
            other_cids = linkx.get_from_ids() + linky.get_from_ids()
        if cen_cids[0] in other_cids or cen_cids[1] in other_cids:
            if set([cen_cids[0]]+[cen_cids[1]]) == set(other_cids):
                return
            else:
                raise ValueError("Cannot display regions if links depend on other components.")
        else:
            self._get_trans_to_cids(cen_cids, other_cids)

    def get_transform_to_cids(self, other_cids):
        """
        Return the function that converts the center components to other_cids.

        We can use this to get the transformation from the x,y coordinates
        that the ExtendedComponent are in to x and y attributes that are
        visualized in a Viewer so that we can translate the geometries
        in the ExtendedComponent to the new coordinates before displaying them.

        Can be called in viewers as:

            >>> tfunc = region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])

        And the function can be used to transform the geometries as:

            >>> from shapely.ops import transform
            >>> new_geoms = [transform(tfunc, g) for g in old_geoms]

        TODO: This is currently hard-coded to work with 2D transformations,
              but could be extended to work with 1D viewers as well. Our region
              geometries are limited to be 2D (or 1D ranges), so links
              that require more than 2 input components do not admit to
              a valid transformation.

        Parameters
        ----------
        other_cid : list of :class:`~glue.core.component.ComponentID`
            The other ComponentIDs (typically the ones that are
            visualized in a Viewer).

        Returns
        -------
        func : `callable`
            The function that converts center_x_id and center_y_id to
            other_cids which can then be used to transform the
            geometries before display. Returns None if there is no
            such valid transformation.
        """

        self.list_of_functions = []
        self._get_trans_to_cids([self.center_x_id, self.center_y_id], other_cids)
        if not self.list_of_functions:
            return None
        elif len(self.list_of_functions) == 1:
            return self.list_of_functions[0]
        else:
            def conv_function(*args):
                # Our list of functions is built up in reverse order
                for f in self.list_of_functions[::-1]:
                    args = f(*args)
                return args
            return conv_function

    def linked_to_center_comp(self, target_cid):
        """
        Check if target_cid can be mapped to one of the center components.

        This is used to see if we can display the ExtendedComponent in a Viewer.

        It is not sufficient to simply see if we can retrieve data from
        target_cid like is commonly done in Viewers:

                >>> _ = self[target_cid]

        Because if target_cid is mapped to a Component that is not one of the
        center components, then we cannot display the regions.

        Parameters
        ----------
        target_cid : :class:`~glue.core.component.ComponentID`
            The ComponentID (typically displayed in a Viewer) which we
            want to check if it is one of the special center components.

        Returns
        -------
        bool
            True if target_cid can be mapped to one of the center components, False otherwise.
        """
        from glue.core.link_manager import is_equivalent_cid  # avoid circular import

        center_cids = [self.center_x_id, self.center_y_id]
        for center_cid in center_cids:
            if is_equivalent_cid(self, center_cid, target_cid):
                return True

        link = self._get_external_link(target_cid)
        if not link:
            return False
        for center_cid in center_cids:
            if center_cid in link:
                return True
            else:
                return any([self.linked_to_center_comp(x) for x in link.get_from_ids()])



import numpy as np
from echo import delay_callback
from glue.tests.visual.helpers import visual_test
from glue.viewers.image.viewer import SimpleImageViewer
from glue.core.application_base import Application
from glue.core.data import Data
from glue.core.link_helpers import LinkSame
from glue.core.data_region import RegionData
from glue.core.data_derived import IndexedData
from astropy.wcs import WCS

from shapely.geometry import Polygon, MultiPolygon, Point
import shapely


@visual_test
def test_simple_image_viewer():

    # Make sure the simple viewer can be instantiated

    data1 = Data(x=np.arange(6).reshape((2, 3)), label='data1')
    data2 = Data(y=2 * np.arange(6).reshape((2, 3)), label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.pixel_component_ids[1] > 1.2)

    return viewer.figure


@visual_test
def test_image_region_layer():
    poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
    poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
    poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
    poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

    polygons = MultiPolygon([poly_3, poly_4])

    geoms = np.array([poly_1, poly_2, polygons])
    a_values = np.array([1, 2, 3])
    b_values = np.array([1, 2, 3])

    region_data = RegionData(regions=geoms, a=a_values, b=b_values)

    image_data = Data(x=np.arange(10000).reshape((100, 100)), label='data1')
    app = Application()
    app.data_collection.append(image_data)
    app.data_collection.append(region_data)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(image_data)
    viewer.add_data(region_data)

    # Link to region data components that are the not the x,y coordinates
    link1 = LinkSame(region_data.id['a'], image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.id['b'], image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    app.data_collection.remove_link(link1)
    app.data_collection.remove_link(link2)

    link1 = LinkSame(region_data.center_x_id, image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.center_y_id, image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    return viewer.figure


def test_region_layer_logic():
    poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
    poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
    poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
    poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

    polygons = MultiPolygon([poly_3, poly_4])

    geoms = np.array([poly_1, poly_2, polygons])
    a_values = np.array([1, 2, 3])
    b_values = np.array([1, 2, 3])

    region_data = RegionData(regions=geoms, a=a_values, b=b_values)

    image_data = Data(x=np.arange(10000).reshape((100, 100)), label='data1')
    app = Application()
    app.data_collection.append(image_data)
    app.data_collection.append(region_data)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(image_data)
    viewer.add_data(region_data)

    assert viewer.layers[0].enabled  # image
    assert not viewer.layers[1].enabled  # regions

    # Link to region data components that are the not the x,y coordinates
    link1 = LinkSame(region_data.id['a'], image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.id['b'], image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    assert viewer.layers[0].enabled  # image
    assert not viewer.layers[1].enabled  # regions

    app.data_collection.remove_link(link1)
    app.data_collection.remove_link(link2)

    link1 = LinkSame(region_data.center_x_id, image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.center_y_id, image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    assert viewer.layers[0].enabled  # image
    assert viewer.layers[1].enabled  # regions


@visual_test
def test_image_region_layer_flip():
    poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
    poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
    poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
    poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

    polygons = MultiPolygon([poly_3, poly_4])

    geoms = np.array([poly_1, poly_2, polygons])
    values = np.array([1, 2, 3])
    region_data = RegionData(regions=geoms, values=values)

    image_data = Data(x=np.arange(10000).reshape((100, 100)), label='data1')
    app = Application()
    app.data_collection.append(image_data)
    app.data_collection.append(region_data)

    link1 = LinkSame(region_data.center_x_id, image_data.pixel_component_ids[0])
    link2 = LinkSame(region_data.center_y_id, image_data.pixel_component_ids[1])
    app.data_collection.add_link(link1)
    app.data_collection.add_link(link2)

    viewer = app.new_data_viewer(SimpleImageViewer)
    viewer.add_data(image_data)
    viewer.add_data(region_data)

    # We need this delay callback here because, while this works in the QT GUI,
    # we need to make sure not to try and redraw the regions while we are flipping
    # the coordinates displayed.

    with delay_callback(viewer.state, 'x_att', 'y_att'):
        viewer.state.x_att = image_data.pixel_component_ids[0]
        viewer.state.y_att = image_data.pixel_component_ids[1]

    return viewer.figure


class TestWCSRegionDisplay(object):
    def setup_method(self, method):

        wcs1 = WCS(naxis=2)
        wcs1.wcs.ctype = 'RA---TAN', 'DEC--TAN'
        wcs1.wcs.crpix = 15, 15
        wcs1.wcs.cd = [[2, -1], [1, 2]]

        wcs1.wcs.set()

        np.random.seed(2)
        self.image1 = Data(label='image1', a=np.random.rand(30, 30), coords=wcs1)
        SHAPELY_ARRAY = np.array([Point(1.5, 2.5).buffer(4),
                                  Polygon([(10, 10), (10, 15), (20, 15), (20, 10)])])
        self.region_data = RegionData(label='My Regions',
                                      color=np.array(['red', 'blue']),
                                      area=shapely.area(SHAPELY_ARRAY),
                                      boundary=SHAPELY_ARRAY)
        self.application = Application()

        self.application.data_collection.append(self.image1)
        self.application.data_collection.append(self.region_data)

        self.viewer = self.application.new_data_viewer(SimpleImageViewer)

    def test_wcs_viewer_bad_link(self):
        self.viewer.add_data(self.image1)

        link1 = LinkSame(self.region_data.id['color'], self.image1.world_component_ids[1])
        link2 = LinkSame(self.region_data.id['area'], self.image1.world_component_ids[0])

        self.application.data_collection.add_link(link1)
        self.application.data_collection.add_link(link2)

        self.viewer.add_data(self.region_data)

        assert self.viewer.state._display_world is True
        assert len(self.viewer.state.layers) == 2
        assert self.viewer.layers[0].enabled
        assert not self.viewer.layers[1].enabled

    def test_wcs_viewer_good_link(self):
        self.viewer.add_data(self.image1)

        link1 = LinkSame(self.region_data.center_x_id, self.image1.world_component_ids[1])
        link2 = LinkSame(self.region_data.center_y_id, self.image1.world_component_ids[0])

        self.application.data_collection.add_link(link1)
        self.application.data_collection.add_link(link2)

        self.viewer.add_data(self.region_data)

        assert self.viewer.state._display_world is True
        assert len(self.viewer.state.layers) == 2
        assert self.viewer.layers[0].enabled
        assert self.viewer.layers[1].enabled

    @visual_test
    def test_image_wcs_viewer(self):
        self.viewer.add_data(self.image1)

        link1 = LinkSame(self.region_data.center_x_id, self.image1.world_component_ids[1])
        link2 = LinkSame(self.region_data.center_y_id, self.image1.world_component_ids[0])

        self.application.data_collection.add_link(link1)
        self.application.data_collection.add_link(link2)

        self.viewer.add_data(self.region_data)

        assert self.viewer.state._display_world is True
        assert len(self.viewer.state.layers) == 2
        assert self.viewer.layers[0].enabled
        assert self.viewer.layers[1].enabled

        return self.viewer.figure

    @visual_test
    def test_image_flipped_wcs_viewer(self):
        self.viewer.add_data(self.image1)

        link1 = LinkSame(self.region_data.center_x_id, self.image1.world_component_ids[1])
        link2 = LinkSame(self.region_data.center_y_id, self.image1.world_component_ids[0])

        self.application.data_collection.add_link(link1)
        self.application.data_collection.add_link(link2)

        self.viewer.add_data(self.region_data)
        original_path_patch = self.viewer.layers[1].region_collection.patches[1].get_path().vertices

        # Flip x,y in the viewer
        with delay_callback(self.viewer.state, 'x_att_world', 'y_att_world', 'x_att', 'y_att'):
            self.viewer.state.x_att_world = self.image1.world_component_ids[0]
            self.viewer.state.y_att_world = self.image1.world_component_ids[1]
            self.viewer.state.x_att = self.image1.pixel_component_ids[0]
            self.viewer.state.y_att = self.image1.pixel_component_ids[1]

        assert self.viewer.state._display_world is True
        assert len(self.viewer.state.layers) == 2
        assert self.viewer.layers[0].enabled
        assert self.viewer.layers[1].enabled
        new_path_patch = self.viewer.layers[1].region_collection.patches[1].get_path().vertices

        # Because we have flipped the viewer, the patches should have changed
        assert np.array_equal(original_path_patch, np.flip(new_path_patch, axis=1))

        return self.viewer.figure


def test_indexed_data():

    # Make sure that the image viewer works properly with IndexedData objects

    data_4d = Data(label='hypercube_wcs',
                   x=np.random.random((3, 5, 4, 3)),
                   coords=WCS(naxis=4))

    data_2d = IndexedData(data_4d, (2, None, 3, None))

    application = Application()

    session = application.session

    hub = session.hub

    data_collection = session.data_collection
    data_collection.append(data_4d)
    data_collection.append(data_2d)

    viewer = application.new_data_viewer(SimpleImageViewer)
    viewer.add_data(data_2d)

    assert viewer.state.x_att is data_2d.pixel_component_ids[1]
    assert viewer.state.y_att is data_2d.pixel_component_ids[0]
    assert viewer.state.x_att_world is data_2d.world_component_ids[1]
    assert viewer.state.y_att_world is data_2d.world_component_ids[0]



import numpy as np
from numpy.testing import assert_allclose

import matplotlib.pyplot as plt

from glue.tests.visual.helpers import visual_test

from glue.viewers.scatter.viewer import SimpleScatterViewer
from glue.core.application_base import Application
from glue.core.data import Data
from glue.core.link_helpers import LinkSame
from glue.core.data_derived import IndexedData


@visual_test
def test_simple_scatter_viewer():

    # Make sure the simple viewer can be instantiated

    data1 = Data(x=[1, 2, 3], label='data1')
    data2 = Data(y=[1, 2, 3], label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleScatterViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.id['x'] > 2)

    return viewer.figure


@visual_test
def test_scatter_density_map():

    # Test the scatter density map

    np.random.seed(12345)

    app = Application()

    x = np.random.normal(3, 1, 100)
    y = np.random.normal(2, 1.5, 100)
    c = np.hypot(x - 3, y - 2)
    s = (x - 3)

    data1 = app.add_data(a={"x": x, "y": y, "c": c, "s": s})[0]

    xx = np.random.normal(3, 1, 1000000)
    yy = np.random.normal(2, 1.5, 1000000)

    data2 = app.add_data(a={"x": xx, "y": yy})[0]

    app.data_collection.add_link(LinkSame(data1.id['x'], data2.id['x']))
    app.data_collection.add_link(LinkSame(data1.id['y'], data2.id['y']))

    viewer = app.new_data_viewer(SimpleScatterViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    viewer.state.layers[0].cmap_mode = 'Linear'
    viewer.state.layers[0].cmap_att = data1.id['c']
    viewer.state.layers[0].cmap = plt.cm.viridis
    viewer.state.layers[0].size_mode = 'Linear'
    viewer.state.layers[0].size_att = data1.id['s']

    viewer.state.layers[1].zorder = 0.5

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.id['x'] > 2)

    return viewer.figure


def test_reset_limits():

    data1 = Data(x=np.arange(1000), y=np.arange(1000) + 1000, label='data')

    app = Application()
    app.data_collection.append(data1)

    viewer = app.new_data_viewer(SimpleScatterViewer)
    viewer.add_data(data1)

    viewer.state.reset_limits()

    # Note that there is a margin included which is why the limits are not 0 to 999

    assert_allclose(viewer.state.x_min, -39.96)
    assert_allclose(viewer.state.x_max, 1038.96)

    assert_allclose(viewer.state.y_min, 1000 - 39.96)
    assert_allclose(viewer.state.y_max, 1000 + 1038.96)

    viewer.state.x_limits_percentile = 90
    viewer.state.y_limits_percentile = 80

    viewer.state.reset_limits()

    assert_allclose(viewer.state.x_min, 13.986)
    assert_allclose(viewer.state.x_max, 985.014)

    assert_allclose(viewer.state.y_min, 1000 + 67.932)
    assert_allclose(viewer.state.y_max, 1000 + 931.068)


def test_indexed_data():

    # Make sure that the scatter viewer works properly with IndexedData objects

    data_4d = Data(label='hypercube',
                   x=np.random.random((3, 5, 4, 3)),
                   y=np.random.random((3, 5, 4, 3)))

    data_2d = IndexedData(data_4d, (2, None, 3, None))

    application = Application()

    session = application.session

    hub = session.hub

    data_collection = session.data_collection
    data_collection.append(data_4d)
    data_collection.append(data_2d)

    viewer = application.new_data_viewer(SimpleScatterViewer)
    viewer.add_data(data_2d)

    assert viewer.state.x_att is data_2d.main_components[0]
    assert viewer.state.y_att is data_2d.main_components[1]



# This artist can be used to deal with the sampling of the data as well as any
# RGB blending.

import warnings

import numpy as np

from glue.config import colormaps, stretches

from matplotlib.colors import ColorConverter, Colormap
from astropy.visualization import ManualInterval, ContrastBiasStretch


__all__ = ['CompositeArray']

COLOR_CONVERTER = ColorConverter()

CMAP_SAMPLING = np.linspace(0, 1, 256)


class CompositeArray(object):

    def __init__(self, **kwargs):

        # We keep a dictionary of layers. The key should be the UUID of the
        # layer artist, and the values should be dictionaries that contain
        # 'zorder', 'visible', 'array', 'color', and 'alpha'.
        self.layers = {}

        self._first = True
        self._mode = 'color'
        self._allow_bad_alpha = False

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        if value not in ['color', 'colormap']:
            raise ValueError("mode should be one of 'color' or 'colormap'")
        self._mode = value

    def allocate(self, uuid):
        self.layers[uuid] = {'zorder': 0,
                             'visible': True,
                             'array': None,
                             'shape': None,
                             'color': '0.5',
                             'cmap': colormaps.members[0][1],
                             'alpha': 1,
                             'clim': (0, 1),
                             'contrast': 1,
                             'bias': 0.5,
                             'stretch': 'linear'}

    def deallocate(self, uuid):
        self.layers.pop(uuid)

    def set(self, uuid, **kwargs):
        for key, value in kwargs.items():
            if key not in self.layers[uuid]:
                raise KeyError("Unknown key: {0}".format(key))
            elif key == 'color' and isinstance(value, Colormap):
                warnings.warn('Setting colormap using "color" key is deprecated, use "cmap" instead.', UserWarning)
                self.layers[uuid]['cmap'] = value
            else:
                self.layers[uuid][key] = value

    @property
    def shape(self):
        for layer in self.layers.values():
            if callable(layer['shape']):
                shape = layer['shape']()
            elif layer['shape'] is not None:
                shape = layer['shape']
            elif callable(layer['array']):
                array = layer['array']()
                if array is None:
                    return None
                else:
                    shape = array.shape
            else:
                shape = layer['array'].shape
            if shape is not None:
                return shape
        return None

    def __getitem__(self, item):
        return self()[item]

    def __call__(self, bounds=None):

        img = None
        visible_layers = 0

        # Get a sorted list of UUIDs with the top layers last
        sorted_uuids = sorted(self.layers, key=lambda x: self.layers[x]['zorder'])

        # We first check that layers are either all colormaps or all single colors.
        # In the case where we are dealing with colormaps, we can start from
        # the last layer that has an opacity of 1 because layers below will not
        # affect the output, assuming also that the colormaps do not change the
        # alpha
        if self.mode == 'colormap':
            for i in range(len(sorted_uuids) - 1, -1, -1):
                layer = self.layers[sorted_uuids[i]]
                if layer['visible']:
                    if layer['alpha'] == 1 and layer['cmap'](CMAP_SAMPLING)[:, 3].min() == 1:
                        sorted_uuids = sorted_uuids[i:]
                        break

        for uuid in sorted_uuids:

            layer = self.layers[uuid]

            if not layer['visible']:
                continue

            interval = ManualInterval(*layer['clim'])
            contrast_bias = ContrastBiasStretch(layer['contrast'], layer['bias'])
            if isinstance(layer['stretch'], str):
                stretch = stretches.members[layer['stretch']]()
            else:
                stretch = layer['stretch']

            if callable(layer['array']):
                array = layer['array'](bounds=bounds)
            else:
                array = layer['array']

            if array is None:
                continue

            if np.isscalar(array):
                scalar = True
                array = np.atleast_2d(array)
            else:
                scalar = False

            data = interval(array)
            data = contrast_bias(data, out=data)
            data = stretch(data, out=data)

            if self.mode == 'colormap':

                # ensure "bad" values have the same alpha as the
                # rest of the layer:
                if hasattr(layer['cmap'], 'get_bad'):
                    bad_rgba = layer['cmap'].get_bad().tolist()
                    bad_color = bad_rgba[:3]
                    bad_alpha = bad_rgba[3:]

                    if self._allow_bad_alpha:
                        bad_rgba = bad_color + bad_alpha
                    else:
                        bad_rgba = bad_color + [layer['alpha']]

                    layer_cmap = layer['cmap'].with_extremes(bad=bad_rgba)
                else:
                    layer_cmap = layer['cmap']

                if img is None:
                    img = np.ones(data.shape + (4,))

                # Compute colormapped image
                plane = layer_cmap(data)

                # Check what the smallest colormap alpha value for this layer is
                # - if it is 1 then this colormap does not change transparency,
                # and this allows us to speed things up a little.
                if layer_cmap(CMAP_SAMPLING)[:, 3].min() == 1 and (not self._allow_bad_alpha or bad_alpha == 1):
                    if layer['alpha'] == 1:
                        img[...] = 0
                    else:
                        plane *= layer['alpha']
                        img *= (1 - layer['alpha'])

                else:

                    # Use traditional alpha compositing
                    alpha_plane = layer['alpha'] * plane[:, :, 3]

                    if self._allow_bad_alpha:
                        # ensure "bad" alpha is preserved:
                        alpha_plane[~np.isfinite(data)] *= bad_alpha

                    plane[:, :, 0] = plane[:, :, 0] * alpha_plane
                    plane[:, :, 1] = plane[:, :, 1] * alpha_plane
                    plane[:, :, 2] = plane[:, :, 2] * alpha_plane

                    img[:, :, 0] *= (1 - alpha_plane)
                    img[:, :, 1] *= (1 - alpha_plane)
                    img[:, :, 2] *= (1 - alpha_plane)

                img[:, :, 3] = 1

            else:
                if img is None:
                    img = np.zeros(data.shape + (4,))

                # Get color and pre-multiply by alpha values
                color = COLOR_CONVERTER.to_rgba_array(layer['color'])[0]
                color *= layer['alpha']

                # We should treat NaN values as zero (post-stretch), which means
                # that those pixels don't contribute towards the final image.
                reset = np.isnan(data)
                if np.any(reset):
                    data[reset] = 0.

                plane = data[:, :, np.newaxis] * color
                plane[:, :, 3] = 1

                visible_layers += 1

            if scalar:
                plane = plane[0, 0]

            img += plane

        if img is None:
            return None
        else:
            img = np.clip(img, 0, 1, out=img)

        return img

    @property
    def dtype(self):
        return np.dtype(float)

    @property
    def ndim(self):
        return 2

    @property
    def size(self):
        return np.prod(self.shape)

    def __contains__(self, item):
        return item in self.layers



# Set up configuration variables

__all__ = ['custom_viewer', 'test']

import os

import sys

import importlib.metadata

__version__ = importlib.metadata.version('glue-core')

from ._mpl_backend import MatplotlibBackendSetter
sys.meta_path.append(MatplotlibBackendSetter())

from glue.viewers.custom.helper import custom_viewer

# Load user's configuration file
from .config import load_configuration
env = load_configuration()

from .main import load_plugins  # noqa


def test(no_optional_skip=False):
    from pytest import main
    root = os.path.abspath(os.path.dirname(__file__))
    args = [root, '-x']
    if no_optional_skip:
        args.append('--no-optional-skip')
    return main(args=args)


from glue._settings_helpers import load_settings
load_settings()



import pytest
from numpy.testing import assert_array_equal

import numpy as np
import shapely
from shapely.geometry import MultiPolygon, Polygon, Point
from shapely.affinity import affine_transform
from shapely.ops import transform

from glue.core.data import Data
from glue.core.data_collection import DataCollection
from glue.core.data_region import RegionData
from glue.core.component import ExtendedComponent
from glue.core.state import GlueUnSerializer
from glue.core.tests.test_application_base import MockApplication
from glue.core.link_helpers import LinkSame, LinkTwoWay
from glue.core.exceptions import IncompatibleAttribute

from glue.plugins.wcs_autolinking.wcs_autolinking import AffineLink

poly_1 = Polygon([(20, 20), (60, 20), (60, 40), (20, 40)])
poly_2 = Polygon([(60, 50), (60, 70), (80, 70), (80, 50)])
poly_3 = Polygon([(10, 10), (15, 10), (15, 15), (10, 15)])
poly_4 = Polygon([(10, 20), (15, 20), (15, 30), (10, 30), (12, 25)])

polygons = MultiPolygon([poly_3, poly_4])
SHAPELY_POLYGON_ARRAY = np.array([poly_1, poly_2, polygons])

SHAPELY_CIRCLE_ARRAY = np.array([Point(0, 0).buffer(1), Point(1, 1).buffer(1)])
CENTER_X = [0, 1]
CENTER_Y = [0, 1]


def shift(x):
    return x + 1


def unshift(x):
    return x - 1


def forwards(x):
    return x * 2


def backwards(x):
    return x / 2


class TestRegionDataLinks(object):
    def setup_method(self):
        self.region_data = RegionData(label='My Regions',
                                      color=np.array(['red', 'blue', 'green']),
                                      area=shapely.area(SHAPELY_POLYGON_ARRAY),
                                      boundary=SHAPELY_POLYGON_ARRAY)
        self.other_data = Data(label='Other Data',
                               x=np.array([1, 2, 3]),
                               y=np.array([5, 6, 7]),
                               color=np.array(['yellow', 'pink', 'orange']),
                               area=np.array([10, 20, 30]))

    def test_linked_properly(self):
        dc = DataCollection([self.region_data, self.other_data])

        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.region_data.center_x_id, self.other_data.id['x']))
        dc.add_link(LinkSame(self.region_data.center_y_id, self.other_data.id['y']))

        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert self.region_data.linked_to_center_comp(viewer_y_att)

    def test_linked_incorrectly(self):
        dc = DataCollection([self.region_data, self.other_data])

        viewer_x_att = self.other_data.id['color']
        viewer_y_att = self.other_data.id['area']

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.region_data.id['color'], self.other_data.id['color']))
        dc.add_link(LinkSame(self.region_data.id['area'], self.other_data.id['area']))

        # This is how a Viewer typically checks to see if is being asked
        # to plot an incompatible dataset. RegionData is a special case
        # because we need to know if the viewer is showing the center components
        # (in which case we can show the regions) or some other components
        # (in which case we can't). In this case we need to know that we
        # CANNOT show the center component.
        try:
            x = self.region_data[viewer_x_att]
        except IncompatibleAttribute:
            assert False

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)


class TestRegionData(object):

    def setup_method(self):
        self.region_data = RegionData(label='My Regions', boundary=SHAPELY_POLYGON_ARRAY)

        self.manual_region_data = RegionData(label='My Manual Regions')

        self.cen_x_id = self.manual_region_data.add_component(CENTER_X, label='Center [x]')
        self.cen_y_id = self.manual_region_data.add_component(CENTER_Y, label='Center [y]')
        self.extended_comp = ExtendedComponent(SHAPELY_CIRCLE_ARRAY, center_comp_ids=[self.cen_x_id, self.cen_y_id])
        self.manual_region_data.add_component(self.extended_comp, label='circles')

        self.other_data = Data(x=np.array([1, 2, 3]), y=np.array([5, 6, 7]), label='Other Data')
        self.mid_data = Data(x=np.array([4, 5, 6]), y=np.array([2, 3, 4]), label='Middle Data')

    def test_basic_properties_simple(self):
        assert self.region_data.label == 'My Regions'
        assert self.region_data.shape == SHAPELY_POLYGON_ARRAY.shape
        assert self.region_data.ndim == 1
        assert self.region_data.size == 3
        assert len(self.region_data.components) == 4
        assert_array_equal(self.region_data['boundary'], SHAPELY_POLYGON_ARRAY)
        assert len(self.region_data.main_components) == 3
        component_labels = [cid.label for cid in self.region_data.main_components]
        assert 'boundary' in component_labels
        assert 'Center [x] for boundary' in component_labels
        assert 'Center [y] for boundary' in component_labels

    def test_basic_properties_manual(self):
        assert self.manual_region_data.label == 'My Manual Regions'
        assert self.manual_region_data.shape == np.asarray(SHAPELY_CIRCLE_ARRAY).shape
        assert self.manual_region_data.ndim == 1
        assert self.manual_region_data.size == 2
        assert len(self.manual_region_data.components) == 4
        assert_array_equal(self.manual_region_data['circles'], SHAPELY_CIRCLE_ARRAY)
        assert len(self.region_data.main_components) == 3
        component_labels = [cid.label for cid in self.manual_region_data.main_components]
        assert 'circles' in component_labels
        assert 'Center [x]' in component_labels
        assert 'Center [y]' in component_labels

    def test_get_kind(self):
        assert self.region_data.get_kind('Center [x] for boundary') == 'numerical'
        assert self.region_data.get_kind('Center [y] for boundary') == 'numerical'
        assert self.region_data.get_kind('boundary') == 'extended'

        assert self.manual_region_data.get_kind('Center [x]') == 'numerical'
        assert self.manual_region_data.get_kind('Center [y]') == 'numerical'
        assert self.manual_region_data.get_kind('circles') == 'extended'

    def test_check_if_can_display(self):
        dc = DataCollection([self.region_data, self.other_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']
        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.region_data.center_x_id, self.other_data.id['x']))
        dc.add_link(LinkSame(self.region_data.center_y_id, self.other_data.id['y']))

        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert self.region_data.linked_to_center_comp(viewer_y_att)

    def test_check_if_can_display_through_intermediate(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.region_data.center_x_id, self.mid_data.id['x']))
        dc.add_link(LinkSame(self.region_data.center_y_id, self.mid_data.id['y']))

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.other_data.id['x'], self.mid_data.id['x']))

        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.other_data.id['y'], self.mid_data.id['y']))

        assert self.region_data.linked_to_center_comp(viewer_y_att)

    def test_check_if_can_display_through_complicated_intermediate(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkSame(self.region_data.center_x_id, self.mid_data.id['x']))
        dc.add_link(LinkSame(self.region_data.center_y_id, self.mid_data.id['y']))

        assert not self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkTwoWay(self.other_data.id['x'], self.mid_data.id['x'], forwards, backwards))

        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert not self.region_data.linked_to_center_comp(viewer_y_att)

        dc.add_link(LinkTwoWay(self.other_data.id['y'], self.mid_data.id['y'], forwards, backwards))

        assert self.region_data.linked_to_center_comp(viewer_y_att)

    def test_get_transformation_to_cid(self):
        dc = DataCollection([self.region_data, self.other_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        dc.add_link(LinkTwoWay(self.region_data.center_x_id, viewer_x_att, forwards, backwards))
        dc.add_link(LinkTwoWay(self.region_data.center_y_id, viewer_y_att, backwards, forwards))

        translation_func = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])
        x_data = self.region_data[self.region_data.center_x_id]
        y_data = self.region_data[self.region_data.center_y_id]

        assert_array_equal(translation_func(x_data, y_data),
                           [forwards(self.region_data[self.region_data.center_x_id]),
                            backwards(self.region_data[self.region_data.center_y_id])])

    def test_get_multilink_transformation(self):
        dc = DataCollection([self.region_data, self.other_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        matrix = np.array([[2, 0, 0], [0, 2, 0], [0, 0, 1]])
        shap_matrix = [2, 0, 0, 2, 0, 0]  # This is how shapely defines this affine matrix
        dc.add_link(AffineLink(self.region_data, self.other_data,
                               [self.region_data.center_x_id, self.region_data.center_y_id],
                               [viewer_x_att, viewer_y_att],
                               matrix=matrix))
        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert self.region_data.linked_to_center_comp(viewer_y_att)

        translation_func = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])
        x_data = self.region_data[self.region_data.center_x_id]
        y_data = self.region_data[self.region_data.center_y_id]
        assert_array_equal(translation_func(x_data, y_data),
                           [self.region_data[viewer_x_att], self.region_data[viewer_y_att]])

        new_regions = [transform(translation_func, g) for g in self.region_data['boundary']]
        shapley_trans = [affine_transform(g, shap_matrix) for g in SHAPELY_POLYGON_ARRAY]
        assert_array_equal(new_regions, shapley_trans)

    def test_get_multilink_transformation_through_int(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        dc.add_link(LinkTwoWay(self.region_data.center_x_id, self.mid_data.id['x'], shift, unshift))
        dc.add_link(LinkTwoWay(self.region_data.center_y_id, self.mid_data.id['y'], unshift, shift))

        matrix = np.array([[2, 0, 0], [0, 3, 0], [0, 0, 1]])

        dc.add_link(AffineLink(self.mid_data, self.other_data,
                               [self.mid_data.id['x'], self.mid_data.id['y']],
                               [viewer_x_att, viewer_y_att],
                               matrix=matrix))
        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert self.region_data.linked_to_center_comp(viewer_y_att)

        translation_func = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])
        x_data = self.region_data[self.region_data.center_x_id]
        y_data = self.region_data[self.region_data.center_y_id]
        assert_array_equal(translation_func(x_data, y_data),
                           [self.region_data[viewer_x_att], self.region_data[viewer_y_att]])

    def test_get_multilink_transformation_through_int_mixed(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']

        dc.add_link(LinkTwoWay(self.region_data.center_x_id, self.mid_data.id['x'], shift, unshift))
        dc.add_link(LinkTwoWay(self.region_data.center_y_id, self.mid_data.id['y'], unshift, shift))

        matrix = np.array([[2, 2, 0], [1, 3, 3], [0, 0, 1]])

        dc.add_link(AffineLink(self.mid_data, self.other_data,
                               [self.mid_data.id['x'], self.mid_data.id['y']],
                               [viewer_x_att, viewer_y_att],
                               matrix=matrix))
        assert self.region_data.linked_to_center_comp(viewer_x_att)
        assert self.region_data.linked_to_center_comp(viewer_y_att)

        translation_func = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])
        x_data = self.region_data[self.region_data.center_x_id]
        y_data = self.region_data[self.region_data.center_y_id]
        assert_array_equal(translation_func(x_data, y_data),
                           [self.region_data[viewer_x_att], self.region_data[viewer_y_att]])

    def test_errors_too_many_viewer_comps(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        self.other_data.add_component(np.array([4, 5, 6]), label='z')
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']
        viewer_z_att = self.other_data.id['z']

        dc.add_link(LinkTwoWay(self.region_data.center_x_id, self.mid_data.id['x'], shift, unshift))
        dc.add_link(LinkTwoWay(self.region_data.center_y_id, self.mid_data.id['y'], unshift, shift))

        with pytest.raises(ValueError, match="Can only deal with 2D transformations"):
            _ = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att, viewer_z_att])

    def test_errors_too_many_comps_in_transform(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        self.other_data.add_component(np.array([4, 5, 6]), label='z')
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']
        viewer_z_att = self.other_data.id['z']
        self.region_data.add_component(np.array([90, 80, 70]), label='zz')

        matrix = np.array([[2, 0, 0, 0], [0, 2, 0, 0], [0, 0, 2, 0], [0, 0, 0, 1]])
        dc.add_link(AffineLink(self.region_data, self.other_data,
                               [self.region_data.center_x_id,
                                self.region_data.center_y_id,
                                self.region_data.id['zz']],
                               [viewer_x_att, viewer_y_att, viewer_z_att],
                               matrix=matrix))

        with pytest.raises(ValueError, match="Can only display regions if links depend on 2 or fewer other components."):
            _ = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])

    def test_errors_bad_link(self):
        dc = DataCollection([self.region_data, self.other_data, self.mid_data])
        self.other_data.add_component(np.array([4, 5, 6]), label='z')
        viewer_x_att = self.other_data.id['x']
        viewer_y_att = self.other_data.id['y']
        viewer_z_att = self.other_data.id['z']
        self.region_data.add_component(np.array([90, 80, 70]), label='zz')
        self.region_data.add_component(np.array([90, 80, 70]), label='fakey')

        matrix = np.array([[2, 0, 0], [0, 2, 0], [0, 0, 1]])
        dc.add_link(AffineLink(self.region_data, self.other_data,
                               [self.region_data.center_x_id,
                                self.region_data.id['fakey']],
                               [viewer_x_att, viewer_y_att],
                               matrix=matrix))

        with pytest.raises(ValueError, match="Cannot display regions if links depend on other components."):
            _ = self.region_data.get_transform_to_cids([viewer_x_att, viewer_y_att])


class TestRegionDataSaveRestore(object):

    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir):
        app = MockApplication()
        geodata = RegionData(label='My Regions', regions=SHAPELY_POLYGON_ARRAY)
        catdata = Data(label='catdata', x=np.array([1, 2, 3, 4]), y=np.array([10, 20, 30, 40]))
        app.data_collection.append(geodata)
        app.data_collection.append(catdata)

        app.data_collection.add_link(LinkSame(geodata.id['Center [x] for regions'], catdata.id['x']))
        app.data_collection.add_link(LinkSame(geodata.id['Center [y] for regions'], catdata.id['y']))

        session_file = tmpdir.mkdir("session").join('test.glu')
        app.save_session(session_file)

        with open(session_file, "r") as f:
            session = f.read()

        state = GlueUnSerializer.loads(session)
        ga = state.object("__main__")
        dc = ga.session.data_collection

        self.reg_before = app.data_collection[0]
        self.cat_before = app.data_collection[1]

        self.reg_after = dc[0]
        self.cat_after = dc[1]

    def test_data_roundtrip(self):
        assert_array_equal(self.reg_before['regions'], self.reg_after['regions'])
        assert_array_equal(self.cat_before['x'], self.cat_after['x'])
        assert_array_equal(self.cat_before['y'], self.cat_after['y'])

    def test_component_ids_are_restored_correctly(self):
        for data in [self.reg_before, self.reg_after]:
            assert data.extended_component_id == data.id['regions']
            assert data.extended_component_id == data.id['regions']

            assert data.components[1] == data.get_component(data.components[3]).x
            assert data.components[2] == data.get_component(data.components[3]).y

    def test_links_still_work(self):
        for data in [(self.reg_before, self.cat_before), (self.reg_after, self.cat_after)]:
            reg_data, cat_data = data
            assert_array_equal(reg_data[reg_data.get_component(reg_data.extended_component_id).x], cat_data.id['x'])
            assert_array_equal(reg_data[reg_data.get_component(reg_data.extended_component_id).y], cat_data.id['y'])



# The following function is a thin wrapper around iter_entry_points. The reason it
# is in this separate file is that when making the Mac app, py2app doesn't
# support entry points, so we replace this function with a version that has the
# entry points we want hardcoded. If this function was in glue/main.py, the
# reference to the iter_plugin_entry_points function in load_plugin would be
# evaluated at compile time rather than at runtime, so the patched version
# wouldn't be used.

import os
import sys
from collections import defaultdict

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points


REQUIRED_PLUGINS = ['glue.plugins.coordinate_helpers',
                    'glue.core.data_exporters',
                    'glue.io.formats.fits']


REQUIRED_PLUGINS_QT = ['glue_qt.plugins.tools.pv_slicer',
                       'glue_qt.viewers.image',
                       'glue_qt.viewers.scatter',
                       'glue_qt.viewers.histogram',
                       'glue_qt.viewers.profile',
                       'glue_qt.viewers.table']


def iter_plugin_entry_points():
    return iter(entry_points(group='glue.plugins'))


class PluginConfig(object):

    def __init__(self, plugins={}):
        self.plugins = defaultdict(lambda: True)
        self.plugins.update(plugins)

    def __str__(self):
        string = ""
        for plugin in sorted(self.plugins):
            string += "{0}: {1}\n".format(plugin, self.plugins[plugin])
        return string

    @classmethod
    def load(cls):

        # Import at runtime because some tests change this value. We also don't
        # just import the variable directly otherwise it is cached.
        from glue import config
        cfg_dir = config.CFG_DIR

        plugin_cfg = os.path.join(cfg_dir, 'plugins.cfg')

        import configparser

        config = configparser.ConfigParser()
        read = config.read(plugin_cfg)

        if len(read) == 0 or not config.has_section('plugins'):
            return cls()

        plugins = {}
        for name, enabled in config.items('plugins'):
            plugins[name] = bool(int(enabled))

        self = cls(plugins=plugins)

        return self

    def save(self):

        # Import at runtime because some tests change this value. We also don't
        # just import the variable directly otherwise it is cached.
        from glue import config
        cfg_dir = config.CFG_DIR

        plugin_cfg = os.path.join(cfg_dir, 'plugins.cfg')

        import configparser

        config = configparser.ConfigParser()
        config.add_section('plugins')

        for key in sorted(self.plugins):
            config.set('plugins', key, value=str(int(self.plugins[key])))

        if not os.path.exists(cfg_dir):
            os.mkdir(cfg_dir)

        with open(plugin_cfg, 'w') as fout:
            config.write(fout)

    def filter(self, keep):
        """
        Keep only certain plugins.

        This is used to filter out plugins that are not installed.
        """
        for key in list(self.plugins.keys())[:]:
            if key not in keep:
                self.plugins.pop(key)



import pytest

import numpy as np
from numpy.testing import assert_equal, assert_allclose

from glue.core import Data, DataCollection
from glue.core.coordinates import Coordinates, IdentityCoordinates
from glue.core.link_helpers import LinkSame
from glue.core.exceptions import IncompatibleDataException, IncompatibleAttribute

from ..state import ImageViewerState, ImageLayerState, AggregateSlice


class SimpleCoordinates(Coordinates):

    def __init__(self):
        super().__init__(pixel_n_dim=3, world_n_dim=3)

    def pixel_to_world_values(self, *args):
        return tuple([2.5 * p for p in args])

    def world_to_pixel_values(self, *args):
        return tuple([0.4 * w for w in args])

    @property
    def axis_correlation_matrix(self):
        matrix = np.zeros((self.world_n_dim, self.pixel_n_dim), dtype=bool)
        matrix[2, 2] = True
        matrix[0:2, 0:2] = True
        return matrix


class TestImageViewerState(object):

    def setup_method(self, method):
        self.state = ImageViewerState()

    def test_pixel_world_linking(self):

        data = Data(label='data', x=[[1, 2], [3, 4]], y=[[5, 6], [7, 8]],
                    coords=IdentityCoordinates(n_dim=2))
        layer_state = ImageLayerState(layer=data, viewer_state=self.state)
        self.state.layers.append(layer_state)

        w1, w2 = data.world_component_ids
        p1, p2 = data.pixel_component_ids

        self.state.reference_data = data

        # Setting world components should set the pixel ones

        self.state.x_att_world = w1
        self.state.y_att_world = w2

        assert self.state.x_att is p1
        assert self.state.y_att is p2

        # Setting one component to the same as the other should trigger the other
        # to flip to prevent them from both being the same

        self.state.x_att_world = w2
        assert self.state.x_att is p2
        assert self.state.y_att is p1
        assert self.state.y_att_world is w1

        self.state.y_att_world = w2
        assert self.state.x_att is p1
        assert self.state.x_att_world is w1
        assert self.state.y_att is p2

        # Changing x_att and y_att should change the world equivalents

        self.state.x_att = p2
        assert self.state.x_att_world is w2
        assert self.state.y_att is p1
        assert self.state.y_att_world is w1

        self.state.y_att = p2
        assert self.state.y_att_world is w2
        assert self.state.x_att is p1
        assert self.state.x_att_world is w1


class TestSlicingAggregation():

    def setup_method(self, method):
        self.viewer_state = ImageViewerState()
        self.data = Data(x=np.ones((3, 4, 5, 6, 7)))
        self.layer_state = ImageLayerState(layer=self.data, viewer_state=self.viewer_state)
        self.viewer_state.layers.append(self.layer_state)
        self.p = self.data.pixel_component_ids

    def test_default(self):
        # Check default settings
        assert self.viewer_state.x_att == self.p[4]
        assert self.viewer_state.y_att == self.p[3]
        assert self.viewer_state.slices == (0, 0, 0, 0, 0)
        assert self.layer_state.get_sliced_data().shape == (6, 7)

    def test_flipped(self):
        # Make sure slice is transposed if needed
        self.viewer_state.x_att = self.p[3]
        self.viewer_state.y_att = self.p[4]
        assert self.viewer_state.slices == (0, 0, 0, 0, 0)
        assert self.layer_state.get_sliced_data().shape == (7, 6)

    def test_slice_preserved(self):
        # Make sure slice stays the same if changing attributes
        self.viewer_state.slices = (1, 3, 2, 5, 4)
        self.viewer_state.x_att = self.p[2]
        self.viewer_state.y_att = self.p[4]
        assert self.viewer_state.slices == (1, 3, 2, 5, 4)
        assert self.viewer_state.wcsaxes_slice == ['y', 5, 'x', 3, 1]
        assert self.layer_state.get_sliced_data().shape == (7, 5)
        self.viewer_state.x_att = self.p[2]
        self.viewer_state.y_att = self.p[1]
        assert self.viewer_state.slices == (1, 3, 2, 5, 4)
        assert self.viewer_state.wcsaxes_slice == [4, 5, 'x', 'y', 1]
        assert self.layer_state.get_sliced_data().shape == (4, 5)
        self.viewer_state.x_att = self.p[0]
        self.viewer_state.y_att = self.p[4]
        assert self.viewer_state.slices == (1, 3, 2, 5, 4)
        assert self.viewer_state.wcsaxes_slice == ['y', 5, 2, 3, 'x']
        assert self.layer_state.get_sliced_data().shape == (7, 3)

    def test_aggregation(self):
        # Check whether using AggregateSlice works
        slc1 = AggregateSlice(slice(None), 0, np.mean)
        slc2 = AggregateSlice(slice(2, 5), 3, np.sum)
        self.viewer_state.slices = (slc1, 3, 2, slc2, 4)
        self.viewer_state.x_att = self.p[2]
        self.viewer_state.y_att = self.p[4]
        assert self.viewer_state.slices == (slc1, 3, 2, slc2, 4)
        assert self.viewer_state.wcsaxes_slice == ['y', 3, 'x', 3, 0]
        result = self.layer_state.get_sliced_data()
        assert result.shape == (7, 5)
        assert_equal(result, 3)  # sum along 3 indices in one of the dimensions


class TestReprojection():

    def setup_method(self, method):

        self.data_collection = DataCollection()

        self.array = np.arange(3024).reshape((6, 7, 8, 9))

        # The reference dataset. Shape is (6, 7, 8, 9).
        self.data1 = Data(x=self.array, coords=IdentityCoordinates(n_dim=4))
        self.data_collection.append(self.data1)

        # A dataset with the same shape but not linked. Shape is (6, 7, 8, 9).
        self.data2 = Data(x=self.array)
        self.data_collection.append(self.data2)

        # A dataset with the same number of dimesnions but in a different
        # order, linked to the first. Shape is (9, 7, 6, 8).
        self.data3 = Data(x=np.moveaxis(self.array, (3, 1, 0, 2), (0, 1, 2, 3)))
        self.data_collection.append(self.data3)
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[0],
                                               self.data3.pixel_component_ids[2]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[1],
                                               self.data3.pixel_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[2],
                                               self.data3.pixel_component_ids[3]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[3],
                                               self.data3.pixel_component_ids[0]))

        # A dataset with fewer dimensions, linked to the first one. Shape is
        # (8, 7, 6)
        self.data4 = Data(x=self.array[:, :, :, 0].transpose())
        self.data_collection.append(self.data4)
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[0],
                                               self.data4.pixel_component_ids[2]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[1],
                                               self.data4.pixel_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[2],
                                               self.data4.pixel_component_ids[0]))

        # A dataset with even fewer dimensions, linked to the first one. Shape
        # is (8, 6)
        self.data5 = Data(x=self.array[:, 0, :, 0].transpose())
        self.data_collection.append(self.data5)
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[0],
                                               self.data5.pixel_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.pixel_component_ids[2],
                                               self.data5.pixel_component_ids[0]))

        # A dataset that is not on the same pixel grid and requires reprojection
        self.data6 = Data()
        self.data6.coords = SimpleCoordinates()
        self.array_nonaligned = np.arange(60).reshape((5, 3, 4))
        self.data6['x'] = np.array(self.array_nonaligned)
        self.data_collection.append(self.data6)
        self.data_collection.add_link(LinkSame(self.data1.world_component_ids[0],
                                               self.data6.world_component_ids[1]))
        self.data_collection.add_link(LinkSame(self.data1.world_component_ids[1],
                                               self.data6.world_component_ids[2]))
        self.data_collection.add_link(LinkSame(self.data1.world_component_ids[2],
                                               self.data6.world_component_ids[0]))

        self.viewer_state = ImageViewerState()
        self.viewer_state.layers.append(ImageLayerState(viewer_state=self.viewer_state, layer=self.data1))
        self.viewer_state.layers.append(ImageLayerState(viewer_state=self.viewer_state, layer=self.data2))
        self.viewer_state.layers.append(ImageLayerState(viewer_state=self.viewer_state, layer=self.data3))
        self.viewer_state.layers.append(ImageLayerState(viewer_state=self.viewer_state, layer=self.data4))
        self.viewer_state.layers.append(ImageLayerState(viewer_state=self.viewer_state, layer=self.data5))
        self.viewer_state.layers.append(ImageLayerState(viewer_state=self.viewer_state, layer=self.data6))

        self.viewer_state.reference_data = self.data1

    def test_default_axis_order(self):

        # Start off with a combination of x/y that means that only one of the
        # other datasets will be matched.

        self.viewer_state.x_att = self.data1.pixel_component_ids[3]
        self.viewer_state.y_att = self.data1.pixel_component_ids[2]
        self.viewer_state.slices = (3, 2, 4, 1)

        image = self.viewer_state.layers[0].get_sliced_data()
        assert_equal(image, self.array[3, 2, :, :])

        with pytest.raises(IncompatibleAttribute):
            self.viewer_state.layers[1].get_sliced_data()

        image = self.viewer_state.layers[2].get_sliced_data()
        assert_equal(image, self.array[3, 2, :, :])

        with pytest.raises(IncompatibleDataException):
            self.viewer_state.layers[3].get_sliced_data()

        with pytest.raises(IncompatibleDataException):
            self.viewer_state.layers[4].get_sliced_data()

    def test_transpose_axis_order(self):

        # Next make it so the x/y axes correspond to the dimensions with length
        # 6 and 8 which most datasets will be compatible with, and this also
        # requires a tranposition.

        self.viewer_state.x_att = self.data1.pixel_component_ids[0]
        self.viewer_state.y_att = self.data1.pixel_component_ids[2]
        self.viewer_state.slices = (3, 2, 4, 1)

        image = self.viewer_state.layers[0].get_sliced_data()
        print(image.shape)
        assert_equal(image, self.array[:, 2, :, 1].transpose())

        with pytest.raises(IncompatibleAttribute):
            self.viewer_state.layers[1].get_sliced_data()

        image = self.viewer_state.layers[2].get_sliced_data()
        print(image.shape)
        assert_equal(image, self.array[:, 2, :, 1].transpose())

        image = self.viewer_state.layers[3].get_sliced_data()
        assert_equal(image, self.array[:, 2, :, 0].transpose())

        image = self.viewer_state.layers[4].get_sliced_data()
        assert_equal(image, self.array[:, 0, :, 0].transpose())

    def test_transpose_axis_order_view(self):

        # As for the previous test, but this time with a view applied

        self.viewer_state.x_att = self.data1.pixel_component_ids[0]
        self.viewer_state.y_att = self.data1.pixel_component_ids[2]
        self.viewer_state.slices = (3, 2, 4, 1)

        view = [slice(1, None, 2), slice(None, None, 3)]

        image = self.viewer_state.layers[0].get_sliced_data(view=view)
        assert_equal(image, self.array[::3, 2, 1::2, 1].transpose())

        with pytest.raises(IncompatibleAttribute):
            self.viewer_state.layers[1].get_sliced_data(view=view)

        image = self.viewer_state.layers[2].get_sliced_data(view=view)
        print(image.shape)
        assert_equal(image, self.array[::3, 2, 1::2, 1].transpose())

        image = self.viewer_state.layers[3].get_sliced_data(view=view)
        assert_equal(image, self.array[::3, 2, 1::2, 0].transpose())

        image = self.viewer_state.layers[4].get_sliced_data(view=view)
        assert_equal(image, self.array[::3, 0, 1::2, 0].transpose())

    def test_reproject(self):

        # Test a case where the data needs to actually be reprojected

        # As for the previous test, but this time with a view applied

        self.viewer_state.x_att = self.data1.pixel_component_ids[0]
        self.viewer_state.y_att = self.data1.pixel_component_ids[2]
        self.viewer_state.slices = (3, 2, 4, 1)

        view = [slice(1, None, 2), slice(None, None, 3)]

        actual = self.viewer_state.layers[5].get_sliced_data(view=view)

        # The data to be reprojected is 3-dimensional. The axes we have set
        # correspond to 1 (for x) and 0 (for y). The third dimension of the
        # data to be reprojected should be sliced. This is linked with the
        # second dimension of the original data, for which the slice index is
        # 2. Since the data to be reprojected has coordinates that are 2.5 times
        # those of the reference data, this means the slice index should be 0.8,
        # which rounded corresponds to 1.
        expected = self.array_nonaligned[:, :, 1]

        # Now in the frame of the reference data, the data to show are indices
        # [0, 3] along x and [1, 3, 5, 7] along y. Applying the transformation,
        # this gives values of [0, 1.2] and [0.4, 1.2, 2, 2.8] for x and y,
        # and rounded, this gives [0, 1] and [0, 1, 2, 3]. As a reminder, in the
        # data to reproject, dimension 0 is y and dimension 1 is x
        expected = expected[:4, :2]

        # Let's make sure this works!
        assert_equal(actual, expected)

    def test_too_many_dimensions(self):

        # If we change the reference data, then the first dataset won't be
        # visible anymore because it has too many dimensions

        self.viewer_state.reference_data = self.data4

        with pytest.raises(IncompatibleAttribute):
            self.viewer_state.layers[0].get_sliced_data()

        self.viewer_state.reference_data = self.data6

        with pytest.raises(IncompatibleAttribute):
            self.viewer_state.layers[0].get_sliced_data()


def test_update_x_att_and_y_att():

    # Regression test for a bug that caused y_att to not be updated before
    # events were sent out about x_att changing.

    viewer_state = ImageViewerState()

    data1 = Data(x=np.ones((3, 4, 5)))

    layer_state1 = ImageLayerState(layer=data1, viewer_state=viewer_state)
    viewer_state.layers.append(layer_state1)

    data2 = Data(x=np.ones((3, 4, 5)))

    layer_state2 = ImageLayerState(layer=data2, viewer_state=viewer_state)
    viewer_state.layers.append(layer_state2)

    def check_consistency(*args, **kwargs):
        # Make sure that x_att and y_att are always for same dataset
        assert viewer_state.x_att.parent is viewer_state.y_att.parent

    viewer_state.add_global_callback(check_consistency)
    viewer_state.add_callback('x_att', check_consistency)
    viewer_state.add_callback('y_att', check_consistency)
    viewer_state.add_callback('x_att_world', check_consistency)
    viewer_state.add_callback('y_att_world', check_consistency)
    viewer_state.add_callback('slices', check_consistency)

    viewer_state.reference_data = data1
    assert viewer_state.x_att is data1.pixel_component_ids[2]
    assert viewer_state.y_att is data1.pixel_component_ids[1]

    viewer_state.reference_data = data2
    assert viewer_state.x_att is data2.pixel_component_ids[2]
    assert viewer_state.y_att is data2.pixel_component_ids[1]


def test_attribute_units():

    # Unit test to make sure that the unit conversion works correctly for
    # v_min/v_max.

    viewer_state = ImageViewerState()

    data1 = Data(x=np.arange(100).reshape((10, 10)))
    data1.get_component('x').units = 'km'

    layer_state1 = ImageLayerState(layer=data1, viewer_state=viewer_state)
    viewer_state.layers.append(layer_state1)

    assert layer_state1.percentile == 100
    assert layer_state1.v_min == 0
    assert layer_state1.v_max == 99

    layer_state1.attribute_display_unit = 'm'

    assert layer_state1.v_min == 0
    assert layer_state1.v_max == 99000

    assert layer_state1.percentile == 100

    layer_state1.percentile = 95

    assert_allclose(layer_state1.v_min, 2475)
    assert_allclose(layer_state1.v_max, 96525)

    assert layer_state1.percentile == 95

    layer_state1.attribute_display_unit = 'km'

    assert_allclose(layer_state1.v_min, 2.475)
    assert_allclose(layer_state1.v_max, 96.525)

    layer_state1.attribute_display_unit = 'm'

    layer_state1.v_max = 50000

    assert layer_state1.percentile == 'Custom'

    layer_state1.attribute_display_unit = 'km'

    assert_allclose(layer_state1.v_min, 2.475)
    assert_allclose(layer_state1.v_max, 50)


def test_stretch_global():

    # Test the option of using global vs per-slice stretch

    viewer_state = ImageViewerState()

    data1 = Data(x=np.arange(1000).reshape((10, 10, 10)))

    layer_state = ImageLayerState(layer=data1, viewer_state=viewer_state)
    viewer_state.layers.append(layer_state)

    assert layer_state.stretch_global is True

    assert layer_state.percentile == 100
    assert layer_state.v_min == 0
    assert layer_state.v_max == 999

    assert viewer_state.slices == (0, 0, 0)

    layer_state.stretch_global = False

    assert layer_state.v_min == 0
    assert layer_state.v_max == 99

    viewer_state.slices = (9, 0, 0)

    assert layer_state.v_min == 900
    assert layer_state.v_max == 999

    layer_state.percentile = 90

    assert layer_state.v_min == 904.95
    assert layer_state.v_max == 994.05

    layer_state.stretch_global = True

    assert layer_state.v_min == 49.95
    assert layer_state.v_max == 949.05


def test_attribute_units_invalid():

    # Regression test for a bug that caused a crash if a dataset had an
    # unrecognized unit

    viewer_state = ImageViewerState()

    data = Data(x=np.arange(100).reshape((10, 10)))
    data.get_component('x').units = 'banana'

    ImageLayerState(layer=data, viewer_state=viewer_state)



import numpy as np

from glue.core.subset import SliceSubsetState
from glue.core.exceptions import IncompatibleAttribute
from glue.core.link_manager import pixel_cid_to_pixel_cid_matrix

__all__ = ['PixelSubsetState']


class PixelSubsetState(SliceSubsetState):

    def copy(self):
        return PixelSubsetState(self.reference_data, self.slices)

    def to_array(self, data, att):

        try:

            return super(PixelSubsetState, self).to_array(data, att)

        except IncompatibleAttribute:

            if data is not self.reference_data:
                pix_coord_out = self._to_linked_pixel_coords(data)
                pix_coord_out = tuple([slice(None) if p is None else slice(p, p + 1) for p in pix_coord_out])
                return data[att, pix_coord_out]

        raise IncompatibleAttribute()

    def _to_linked_pixel_coords(self, data):

        # Determine which pixel dimensions are being sliced over
        dimensions = [idim for idim, slc in enumerate(self.slices) if slc.start is not None]

        # Determine pixel to pixel correlation matrix
        matrix = pixel_cid_to_pixel_cid_matrix(self.reference_data, data)

        # Find pixel dimensions in 'data' that are correlated
        correlated_dims = np.nonzero(np.any(matrix[dimensions], axis=0))[0]

        # Check that if we do the operation backwards we just get the
        # original dimensions back
        check_dimensions = np.nonzero(np.any(matrix[:, correlated_dims], axis=1))[0]

        if np.array_equal(dimensions, check_dimensions):
            pix_coord_in = tuple([slice(0, 1) if slc.start is None else slc for slc in self.slices])
            pix_coord_out = []
            for idim, pix_cid in enumerate(data.pixel_component_ids):
                if idim in correlated_dims:
                    coord = int(np.round(self.reference_data[pix_cid, pix_coord_in].ravel()[0]))
                else:
                    coord = None
                pix_coord_out.append(coord)

            return pix_coord_out

        raise IncompatibleAttribute()

    def get_xy(self, data, dim1, dim2):
        pix_coord_out = self._to_linked_pixel_coords(data)
        if pix_coord_out[dim1] is None or pix_coord_out[dim2] is None:
            raise IncompatibleAttribute
        else:
            return pix_coord_out[dim1], pix_coord_out[dim2]



#!/usr/bin/env python

from importlib import import_module

from glue.logger import logger
from glue._plugin_helpers import REQUIRED_PLUGINS, REQUIRED_PLUGINS_QT


_loaded_plugins = set()
_installed_plugins = set()


def load_plugins(splash=None, require_qt_plugins=False, plugins_to_load=None):
    """

    Parameters
    ----------
    splash : default: None
        instance of splash http rendering service
    require_qt_plugins : boolean default: False
        whether to use qt plugins defined in constant REQUIRED_PLUGINS_QT
    plugins_to_load : list
        desired valid plugin strings

    Returns
    -------

    """

    # Search for plugins installed via entry_points. Basically, any package can
    # define plugins for glue, and needs to define an entry point using the
    # following format:
    #
    # entry_points = """
    # [glue.plugins]
    # webcam_importer=glue_exp.importers.webcam:setup
    # vizier_importer=glue_exp.importers.vizier:setup
    # dataverse_importer=glue_exp.importers.dataverse:setup
    # """
    #
    # where ``setup`` is a function that does whatever is needed to set up the
    # plugin, such as add items to various registries.

    import setuptools
    logger.info("Loading external plugins using "
                "setuptools=={0}".format(setuptools.__version__))

    from glue._plugin_helpers import iter_plugin_entry_points, PluginConfig
    config = PluginConfig.load()

    if plugins_to_load is None:
        plugins_to_load = [i.module for i in list(iter_plugin_entry_points())]
        if require_qt_plugins:
            plugins_to_require = [*REQUIRED_PLUGINS, *REQUIRED_PLUGINS_QT]
        else:
            plugins_to_require = REQUIRED_PLUGINS
    else:
        plugins_to_require = plugins_to_load
    n_plugins = len(plugins_to_require)

    for i_plugin, item in enumerate(list(iter_plugin_entry_points())):

        if item.module in plugins_to_load:
            if item.module not in _installed_plugins:
                _installed_plugins.add(item.name)

            if item.module in _loaded_plugins:
                logger.info("Plugin {0} already loaded".format(item.name))
                continue

            # loads all plugins, want to make this more customisable
            if not config.plugins[item.name]:
                continue

        # We don't use item.load() because that then checks requirements of all
        # the imported packages, which can lead to errors like this one that
        # don't really matter:
        #
        # Exception: (pytest 2.6.0 (/Users/tom/miniconda3/envs/py27/lib/python2.7/site-packages),
        #             Requirement.parse('pytest>=2.8'), set(['astropy']))
        #
        # Just to be clear, this kind of error does indicate that there is an
        # old version of a package in the environment, but this can confuse
        # users as importing astropy directly would work (as setuptools then
        # doesn't do a stringent test of dependency versions). Often this kind
        # of error can occur if there is a conda version of a package and an
        # older pip version.

        try:
            module = import_module(item.module)
            function = getattr(module, item.attr)
            function()
        except Exception as exc:
            # Here we check that some of the 'core' plugins load well and
            # raise an actual exception if not.
            if item.module in plugins_to_require:
                raise
            else:
                logger.info("Loading plugin {0} failed "
                            "(Exception: {1})".format(item.name, exc))
        else:
            logger.info("Loading plugin {0} succeeded".format(item.name))
            _loaded_plugins.add(item.module)

        if splash is not None:
            splash.set_progress(100. * i_plugin / float(n_plugins))

    try:
        config.save()
    except Exception as e:
        logger.warn("Failed to load plugin configuration")

    # Reload the settings now that we have loaded plugins, since some plugins
    # may have added some settings. Note that this will not re-read settings
    # that were previously read.
    from glue._settings_helpers import load_settings
    load_settings()



import os
import mmap
import warnings
from collections import OrderedDict

import numpy as np

from glue.core.data import Component, Data
from glue.config import data_factory


__all__ = ['is_hdf5', 'hdf5_reader']


def mmap_info_to_array(info, mapping):
    shape = info['shape']
    dtype = info['dtype']
    offset = info['offset']
    length = np.prod(shape)
    return np.frombuffer(mapping, dtype=dtype, count=length, offset=offset).reshape(shape)


def extract_hdf5_datasets(filename, memmap=True):
    """
    Recursive function that returns a dictionary with all the datasets found in
    an HDF5 file or group. `handle` should be an instance of h5py.highlevel.File
    or h5py.highlevel.Group.

    Parameters
    ----------
    filename : str or file-like
        The path or file handle to the HDF5 file
    memmap : bool, optional
        Whether to use memory mapping
    """

    import h5py

    file_handle = h5py.File(filename, 'r')

    arrays = {}

    from astropy.table import Table

    def visitor(name, item):
        if isinstance(item, h5py.Dataset):
            full_path = item.name
            if item.dtype.kind in ('f', 'i', 'S'):
                offset = item.id.get_offset()
                # If an offset is available, the data is contiguous and we can
                # use memory mapping for efficiency.
                if not memmap or offset is None:
                    arrays[full_path] = item[()]
                else:
                    arrays[full_path] = dict(offset=offset, shape=item.shape, dtype=item.dtype)
            elif item.dtype.kind in ('V',):
                arrays[full_path] = Table.read(item, format='hdf5')

    file_handle.visititems(visitor)
    file_handle.close()

    # Now create memory-mapped arrays

    with open(filename, 'rb') as data_file:
        fileno = data_file.fileno()
        mapping = mmap.mmap(fileno, 0, access=mmap.ACCESS_READ)
        for name, value in arrays.items():
            if isinstance(value, dict):
                arrays[name] = mmap_info_to_array(value, mapping)

    return arrays


def is_hdf5(filename):
    # All hdf5 files begin with the same sequence
    with open(filename, 'rb') as infile:
        return infile.read(8) == b'\x89HDF\r\n\x1a\n'


@data_factory(label="HDF5 file", identifier=is_hdf5, priority=100)
def hdf5_reader(filename, auto_merge=True, memmap=True, **kwargs):
    """
    Read in all datasets from an HDF5 file

    Parameters
    ----------
    filename : str or file-like
        The path or file handle to the HDF5 file
    auto_merge : bool
        If all datasets have the same shape, and are at the base of the file,
        assume they are a column-based table and merge them into a single dataset.
    memmap : bool, optional
        Whether to use memory mapping
    """

    from astropy.table import Table

    # Read in all datasets
    datasets = extract_hdf5_datasets(filename, memmap=memmap)

    label_base = os.path.basename(filename).rpartition('.')[0]

    if not label_base:
        label_base = os.path.basename(filename)

    if len(datasets) == 0:
        return

    if not auto_merge or len(datasets) == 1 or any([isinstance(data, Table) for data in datasets.values()]):
        merge_data = False
    else:
        reference_shape = list(datasets.values())[0].shape
        merge_data = all([data.shape == reference_shape and key.count('/') == 1 for key, data in datasets.items()])

    groups = OrderedDict()

    data = None

    for key in datasets:
        label = '{0}[{1}]'.format(label_base, key)
        array = datasets[key]
        if isinstance(array, Table):
            data = Data(label=label)
            groups[label] = data
            for column_name in array.columns:
                column = array[column_name]
                if column.ndim == 1:
                    component = Component.autotyped(column, units=column.unit)
                    data.add_component(component=component,
                                       label=column_name)
                else:
                    warnings.warn("HDF5: Ignoring vector column {0}".format(column_name))
        else:
            if data is None and merge_data:
                data = Data(label=label_base)
                groups[label_base] = data
            elif not merge_data:
                data = Data(label=label)
                groups[label] = data
            data.add_component(component=datasets[key], label=key[1:])

    return [groups[idx] for idx in sorted(groups)]



from glue.core.data_factories.helpers import has_extension  # noqa


__all__ = []

try:
    from glue.core.data_factories.dendro_loader import load_dendro, is_dendro  # noqa
except ImportError:
    pass



from astropy import units as u
from astropy.wcs import WCS

import numpy as np
from numpy.testing import assert_equal, assert_allclose

from glue.tests.visual.helpers import visual_test
from glue.viewers.profile.viewer import SimpleProfileViewer
from glue.core.application_base import Application
from glue.core.data import Data
from glue.config import settings, unit_converter
from glue.plugins.wcs_autolinking.wcs_autolinking import WCSLink
from glue.core.roi import XRangeROI
from glue.core.data_derived import IndexedData


@visual_test
def test_simple_profile_viewer():

    # Make sure the simple viewer can be instantiated

    data1 = Data(x=[1, 2, 3], label='data1')
    data2 = Data(y=[1, 2, 3], label='data2')

    app = Application()
    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(SimpleProfileViewer)
    viewer.add_data(data1)
    viewer.add_data(data2)

    app.data_collection.new_subset_group(label='subset1', subset_state=data1.pixel_component_ids[0] > 0.8)

    viewer.state.layers[2].linewidth = 5

    return viewer.figure


@unit_converter('test-spectral2')
class SpectralUnitConverter:

    def equivalent_units(self, data, cid, units):
        return map(str, u.Unit(units).find_equivalent_units(include_prefix_units=True, equivalencies=u.spectral()))

    def to_unit(self, data, cid, values, original_units, target_units):
        return (values * u.Unit(original_units)).to_value(target_units, equivalencies=u.spectral())


def test_unit_conversion():

    settings.UNIT_CONVERTER = 'test-spectral2'

    wcs1 = WCS(naxis=1)
    wcs1.wcs.ctype = ['FREQ']
    wcs1.wcs.crval = [1]
    wcs1.wcs.cdelt = [1]
    wcs1.wcs.crpix = [1]
    wcs1.wcs.cunit = ['GHz']

    d1 = Data(f1=[1, 2, 3])
    d1.get_component('f1').units = 'Jy'
    d1.coords = wcs1

    wcs2 = WCS(naxis=1)
    wcs2.wcs.ctype = ['WAVE']
    wcs2.wcs.crval = [10]
    wcs2.wcs.cdelt = [10]
    wcs2.wcs.crpix = [1]
    wcs2.wcs.cunit = ['cm']

    d2 = Data(f2=[2000, 1000, 3000])
    d2.get_component('f2').units = 'mJy'
    d2.coords = wcs2

    app = Application()
    session = app.session

    data_collection = session.data_collection
    data_collection.append(d1)
    data_collection.append(d2)

    data_collection.add_link(WCSLink(d1, d2))

    viewer = app.new_data_viewer(SimpleProfileViewer)
    viewer.add_data(d1)
    viewer.add_data(d2)

    assert viewer.layers[0].enabled
    assert viewer.layers[1].enabled

    x, y = viewer.state.layers[0].profile
    assert_allclose(x, [1.e9, 2.e9, 3.e9])
    assert_allclose(y, [1, 2, 3])

    x, y = viewer.state.layers[1].profile
    assert_allclose(x, 299792458 / np.array([0.1, 0.2, 0.3]))
    assert_allclose(y, [2000, 1000, 3000])

    assert viewer.state.x_min == 1.e9
    assert viewer.state.x_max == 3.e9
    assert viewer.state.y_min == 1.
    assert viewer.state.y_max == 3.

    # Change the limits to make sure they are always converted
    viewer.state.x_min = 5e8
    viewer.state.x_max = 4e9
    viewer.state.y_min = 0.5
    viewer.state.y_max = 3.5

    roi = XRangeROI(1.4e9, 2.1e9)
    viewer.apply_roi(roi)

    assert len(d1.subsets) == 1
    assert_equal(d1.subsets[0].to_mask(), [0, 1, 0])

    assert len(d2.subsets) == 1
    assert_equal(d2.subsets[0].to_mask(), [0, 1, 0])

    viewer.state.x_display_unit = 'GHz'
    viewer.state.y_display_unit = 'mJy'

    x, y = viewer.state.layers[0].profile
    assert_allclose(x, [1, 2, 3])
    assert_allclose(y, [1000, 2000, 3000])

    x, y = viewer.state.layers[1].profile
    assert_allclose(x, 2.99792458 / np.array([1, 2, 3]))
    assert_allclose(y, [2000, 1000, 3000])

    assert viewer.state.x_min == 0.5
    assert viewer.state.x_max == 4.

    # Units get reset because they were originally 'native' and 'native' to a
    # specific unit always trigger resetting the limits since different datasets
    # might be converted in different ways.
    assert viewer.state.y_min == 1000.
    assert viewer.state.y_max == 3000.

    # Now set the limits explicitly again and make sure in future they are converted
    viewer.state.y_min = 500.
    viewer.state.y_max = 3500.

    roi = XRangeROI(0.5, 1.2)
    viewer.apply_roi(roi)

    assert len(d1.subsets) == 1
    assert_equal(d1.subsets[0].to_mask(), [1, 0, 0])

    assert len(d2.subsets) == 1
    assert_equal(d2.subsets[0].to_mask(), [0, 0, 1])

    viewer.state.x_display_unit = 'cm'
    viewer.state.y_display_unit = 'Jy'

    roi = XRangeROI(15, 35)
    viewer.apply_roi(roi)

    assert len(d1.subsets) == 1
    assert_equal(d1.subsets[0].to_mask(), [1, 0, 0])

    assert len(d2.subsets) == 1
    assert_equal(d2.subsets[0].to_mask(), [0, 1, 1])

    assert_allclose(viewer.state.x_min, (4 * u.GHz).to_value(u.cm, equivalencies=u.spectral()))
    assert_allclose(viewer.state.x_max, (0.5 * u.GHz).to_value(u.cm, equivalencies=u.spectral()))
    assert_allclose(viewer.state.y_min, 0.5)
    assert_allclose(viewer.state.y_max, 3.5)

    # Regression test for a bug that caused unit changes to not work on y axis
    # if reference data was not first layer

    viewer.state.reference_data = d2
    viewer.state.y_display_unit = 'mJy'


def test_indexed_data():

    # Make sure that the profile viewer works properly with IndexedData objects

    data_4d = Data(label='hypercube_wcs',
                   x=np.random.random((3, 5, 4, 3)),
                   coords=WCS(naxis=4))

    data_2d = IndexedData(data_4d, (2, None, 3, None))

    application = Application()

    session = application.session

    hub = session.hub

    data_collection = session.data_collection
    data_collection.append(data_4d)
    data_collection.append(data_2d)

    viewer = application.new_data_viewer(SimpleProfileViewer)
    viewer.add_data(data_2d)

    assert viewer.state.x_att is data_2d.world_component_ids[0]



import numpy as np

from numpy.testing import assert_allclose

from echo import CallbackProperty, ListCallbackProperty
from glue.core import Data, DataCollection

from .test_state import clone
from ..state_objects import (State, StateAttributeLimitsHelper,
                             StateAttributeSingleValueHelper,
                             StateAttributeHistogramHelper)


class SimpleTestState(State):
    a = CallbackProperty()
    b = CallbackProperty()
    flat = ListCallbackProperty()
    nested = ListCallbackProperty()


def test_state_serialization():

    state1 = SimpleTestState()
    state1.a = 2
    state1.b = 'hello'
    state1.flat = [1, 3, 4]

    sub_state = SimpleTestState()
    sub_state.a = 3
    sub_state.b = 'blah'
    sub_state.flat = [1, 2]
    sub_state.nested = []

    state1.nested = [1, 3, sub_state]

    state2 = clone(state1)

    assert state2.a == 2
    assert state2.b == 'hello'
    assert state2.flat == [1, 3, 4]
    assert state2.nested[0:2] == [1, 3]
    assert state2.nested[2].a == 3
    assert state2.nested[2].b == 'blah'
    assert state2.nested[2].flat == [1, 2]
    assert state2.nested[2].nested == []


EXPECTED_STR = """
a: 2
b: hello
flat: <CallbackList with 3 elements>
nested: <CallbackList with 3 elements>
"""

EXPECTED_REPR = """
<SimpleTestState
  a: 2
  b: hello
  flat: <CallbackList with 3 elements>
  nested: <CallbackList with 3 elements>
>
"""


def test_state_str_repr():

    state1 = SimpleTestState()
    state1.a = 2
    state1.b = 'hello'
    state1.flat = [1, 3, 4]

    sub_state = SimpleTestState()

    state1.nested = [1, 3, sub_state]

    assert str(state1) == EXPECTED_STR.strip()
    assert repr(state1) == EXPECTED_REPR.strip()


class TestStateAttributeLimitsHelper():

    def setup_method(self, method):

        self.data = Data(x=np.linspace(-100, 100, 10000),
                         y=np.linspace(2, 3, 10000), label='test_data')

        self.data_collection = DataCollection([self.data])

        class SimpleState(State):

            layer = CallbackProperty()
            comp = CallbackProperty()
            lower = CallbackProperty()
            upper = CallbackProperty()
            log = CallbackProperty(False)
            scale = CallbackProperty(100)

        self.state = SimpleState()

        self.helper = StateAttributeLimitsHelper(self.state, attribute='comp',
                                                 lower='lower', upper='upper',
                                                 percentile='scale', log='log')
        self.state.data = self.data
        self.state.comp = self.data.id['x']

        self.x_id = self.data.main_components[0]
        self.y_id = self.data.main_components[1]

    def test_minmax(self):
        assert self.helper.lower == -100
        assert self.helper.upper == +100

    def test_change_attribute(self):
        self.helper.attribute = self.y_id
        assert self.helper.lower == 2
        assert self.helper.upper == 3
        self.helper.attribute = self.x_id
        assert self.helper.lower == -100
        assert self.helper.upper == +100

    def test_change_percentile(self):

        # Changing scale mode updates the limits
        self.helper.percentile = 99.5
        assert_allclose(self.helper.lower, -99.5)
        assert_allclose(self.helper.upper, +99.5)
        self.helper.percentile = 99
        assert_allclose(self.helper.lower, -99)
        assert_allclose(self.helper.upper, +99)
        self.helper.percentile = 90
        assert_allclose(self.helper.lower, -90)
        assert_allclose(self.helper.upper, +90)

        # When switching to custom, the last limits are retained
        self.helper.percentile = "Custom"
        assert_allclose(self.helper.lower, -90)
        assert_allclose(self.helper.upper, +90)

    def test_percentile_cached(self):
        # Make sure that if we change scale and change attribute, the scale
        # modes are cached on a per-attribute basis.
        self.helper.percentile = 99.5
        self.state.comp = self.y_id
        assert self.helper.percentile == 100
        self.helper.percentile = 99
        self.state.comp = self.x_id
        assert self.helper.percentile == 99.5
        self.state.comp = self.y_id
        assert self.helper.percentile == 99

    def test_flip_button(self):

        self.helper.flip_limits()

        assert self.helper.lower == +100
        assert self.helper.upper == -100

        # Make sure that values were re-cached when flipping
        self.state.comp = self.y_id
        assert self.helper.lower == 2
        assert self.helper.upper == 3
        self.state.comp = self.x_id
        assert self.helper.lower == +100
        assert self.helper.upper == -100

    def test_manual_edit(self):

        # Make sure that values are re-cached when edited manually
        self.helper.percentile = "Custom"
        self.state.lower = -122
        self.state.upper = 234
        self.helper.log = True
        assert self.helper.lower == -122
        assert self.helper.upper == 234
        assert self.helper.log
        self.state.comp = self.y_id
        assert self.helper.lower == 2
        assert self.helper.upper == 3
        assert not self.helper.log
        self.state.comp = self.x_id
        assert self.helper.lower == -122
        assert self.helper.upper == 234
        assert self.helper.log

    def test_set_slice(self):

        # Set subset to compute limits from slice
        self.helper.set_slice([slice(2000, 8000)])

        assert self.helper.percentile == 100

        assert_allclose(self.helper.lower, -59.996)
        assert_allclose(self.helper.upper, 59.996)

        self.helper.percentile = 90

        assert_allclose(self.helper.lower, -53.9964)
        assert_allclose(self.helper.upper, 53.9964)

        self.helper.set_slice(None)

        self.helper.percentile = 95

        assert_allclose(self.helper.lower, -95)
        assert_allclose(self.helper.upper, 95)


class TestStateAttributeSingleValueHelper():

    def setup_method(self, method):

        self.data = Data(x=np.linspace(-100, 30, 9999),
                         y=np.linspace(2, 3, 9999), label='test_data')

        self.data_collection = DataCollection([self.data])

        class SimpleState(State):

            layer = CallbackProperty()
            comp = CallbackProperty()
            val = CallbackProperty()

        self.state = SimpleState()

        self.helper = StateAttributeSingleValueHelper(self.state, attribute='comp',
                                                      function=np.nanmedian, value='val')

        self.state.data = self.data

        self.state.comp = self.data.id['x']

        self.x_id = self.data.main_components[0]
        self.y_id = self.data.main_components[1]

    def test_value(self):
        assert self.helper.value == -35.

    def test_change_attribute(self):
        self.helper.attribute = self.y_id
        assert self.helper.value == 2.5
        self.helper.attribute = self.x_id
        assert self.helper.value == -35

    def test_manual_edit(self):
        self.state.val = 42.
        assert self.helper.value == 42
        self.state.comp = self.y_id
        assert self.helper.value == 2.5
        self.state.comp = self.x_id
        assert self.helper.value == 42


class TestStateAttributeHistogramHelper():

    def setup_method(self, method):

        self.data = Data(x=[-3.2, 4.3, 2.2, 5.4, 7.2, -1.1, 2.3],
                         y=['a', 'f', 'd', 'e', 'f', 'f', 'a'], label='test_data')

        self.data_collection = DataCollection([self.data])

        class SimpleState(State):

            layer = CallbackProperty()
            comp = CallbackProperty()
            x_min = CallbackProperty()
            x_max = CallbackProperty()
            n_bin = CallbackProperty()

        self.state = SimpleState()

        self.helper = StateAttributeHistogramHelper(self.state, attribute='comp',
                                                    lower='x_min', upper='x_max', n_bin='n_bin')

        self.state.data = self.data

    def test_default_numerical(self):
        self.state.comp = self.data.id['x']
        assert self.state.x_min == -3.2
        assert self.state.x_max == 7.2
        assert self.state.n_bin == 15

    def test_default_categorical(self):
        self.state.comp = self.data.id['y']
        assert self.state.x_min == -0.5
        assert self.state.x_max == 3.5
        assert self.state.n_bin == 4

    def test_hitting_limits(self):

        # FIXME: here we modify the internal defaults rather than making a new
        # state helper, but this could be improved
        self.helper._default_n_bin = 4
        self.helper._max_n_bin = 3

        self.state.comp = self.data.id['x']
        assert self.state.x_min == -3.2
        assert self.state.x_max == 7.2
        assert self.state.n_bin == 4

        self.state.comp = self.data.id['y']
        assert self.state.x_min == -0.5
        assert self.state.x_max == 3.5
        assert self.state.n_bin == 3

    def test_caching(self):
        self.state.comp = self.data.id['x']
        self.state.x_min = 2
        self.state.x_max = 7
        self.state.n_bin = 8
        self.state.comp = self.data.id['y']
        self.state.x_min = 1.5
        self.state.x_max = 3.5
        self.state.n_bin = 3
        self.state.comp = self.data.id['x']
        assert self.state.x_min == 2
        assert self.state.x_max == 7
        assert self.state.n_bin == 8
        self.state.comp = self.data.id['y']
        assert self.state.x_min == 1.5
        assert self.state.x_max == 3.5
        assert self.state.n_bin == 3


def test_histogram_helper_common_n_bin():

    data = Data(x=[-3.2, 4.3, 2.2],
                y=['a', 'f', 'd'],
                z=[1.1, 2.3, 1.2],
                label='test_data')

    class SimpleState(State):

        layer = CallbackProperty()
        comp = CallbackProperty()
        x_min = CallbackProperty()
        x_max = CallbackProperty()
        n_bin = CallbackProperty()
        common = CallbackProperty()

    state = SimpleState()

    helper = StateAttributeHistogramHelper(state, attribute='comp',
                                           lower='x_min', upper='x_max', n_bin='n_bin',
                                           common_n_bin='common')

    state.data = data

    state.comp = data.id['x']
    state.n_bin = 9
    state.comp = data.id['y']
    assert state.n_bin == 3
    state.comp = data.id['z']
    assert state.n_bin == 15

    state.n_bin = 12

    state.common = True

    state.comp = data.id['x']
    assert state.n_bin == 12

    state.n_bin = 11

    state.comp = data.id['y']
    assert state.n_bin == 3
    state.comp = data.id['z']
    assert state.n_bin == 11

    state.common = False
    state.n_bin = 13

    state.comp = data.id['x']
    assert state.n_bin == 11


def test_histogram_helper_common_n_bin_active():

    # Make sure that common_n_bin works as expected if True from start

    data = Data(x=[-3.2, 4.3, 2.2],
                y=['a', 'f', 'd'],
                z=[1.1, 2.3, 1.2],
                label='test_data')

    class SimpleState(State):

        layer = CallbackProperty()
        comp = CallbackProperty()
        x_min = CallbackProperty()
        x_max = CallbackProperty()
        n_bin = CallbackProperty()
        common = CallbackProperty(True)

    state = SimpleState()

    helper = StateAttributeHistogramHelper(state, attribute='comp',
                                           lower='x_min', upper='x_max', n_bin='n_bin',
                                           common_n_bin='common')

    state.data = data

    state.comp = data.id['x']
    state.n_bin = 9
    state.comp = data.id['z']
    assert state.n_bin == 9

    state.n_bin = 12

    state.common = True

    state.comp = data.id['x']
    assert state.n_bin == 12

    state.n_bin = 11

    state.comp = data.id['y']
    assert state.n_bin == 3
    state.comp = data.id['z']
    assert state.n_bin == 11

    state.common = False
    state.n_bin = 13

    state.comp = data.id['x']
    assert state.n_bin == 11


def test_limits_helper_initial_values():

    # Regression test for a bug that occurred if the limits cache was empty
    # but some attributes were set to values - in this case we don't want to
    # override the existing values.

    data = Data(x=np.linspace(-100, 100, 10000),
                y=np.linspace(2, 3, 10000), label='test_data')

    class SimpleState(State):

        layer = CallbackProperty()
        comp = CallbackProperty()
        lower = CallbackProperty()
        upper = CallbackProperty()

    state = SimpleState()
    state.lower = 1
    state.upper = 2
    state.comp = data.id['x']

    helper = StateAttributeLimitsHelper(state, attribute='comp',
                                        lower='lower', upper='upper')

    assert helper.lower == 1
    assert helper.upper == 2


class DatetimeState(State):
    a = CallbackProperty()


def test_state_serialization_datetime64():

    state1 = DatetimeState()
    state1.a = np.datetime64(100, 'D')

    state2 = clone(state1)

    assert state2.a == np.datetime64(100, 'D')


def test_nan_inf_minmax():

    data = Data(x=[3, 1, -2, np.inf, np.nan], label='test_data')

    class SimpleState(State):

        layer = CallbackProperty()
        comp = CallbackProperty()
        lower = CallbackProperty()
        upper = CallbackProperty()
        percentile = CallbackProperty()
        log = CallbackProperty()

    state = SimpleState()

    helper = StateAttributeLimitsHelper(state, attribute='comp',  # noqa
                                        lower='lower', upper='upper',
                                        percentile='percentile', log='log')

    state.data = data
    state.comp = data.id['x']

    assert state.lower == -2
    assert state.upper == +3

    state.log = True

    assert state.lower == +1
    assert state.upper == +3

    state.log = False
    state.percentile = 99

    assert_allclose(state.lower, -1.97)
    assert_allclose(state.upper, +2.98)


def test_percentile_no_log():

    # Regression test for a bug that caused a crash if the state class had a
    # percentile attribute but no log.

    data = Data(x=np.linspace(-100, 100, 10000),
                y=np.linspace(2, 3, 10000), label='test_data')

    class SimpleState(State):

        layer = CallbackProperty()
        comp = CallbackProperty()
        lower = CallbackProperty()
        upper = CallbackProperty()
        scale = CallbackProperty()

    state = SimpleState()

    state.comp = data.id['x']
    state.lower = 2
    state.upper = 4

    helper = StateAttributeLimitsHelper(state, attribute='comp',
                                        lower='lower', upper='upper',
                                        percentile='scale')

    state.scale = 90
